from copy import deepcopy
from collections import defaultdict

import numpy as np

from sklearn import clone
from sklearn.preprocessing import OneHotEncoder, OrdinalEncoder
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import KFold

from .base_models import BaseEstModel
from .utils import (
    convert2array,
    convert4onehot,
    nd_kron,
    get_wv,
    cartesian,
    get_tr_ctrl,
)
from ylearn.utils import logging

logger = logging.get_logger(__name__)

#
# class DoubleML(BaseEstModel):
#     r"""
#     Double machine learning has two stages:
#     In stage I, we
#         1. fit a model (y_model) to predict outcome (y) from confounders (w) to
#             get the predicted outcome (py);
#         2. fit a model (x_model) to predict treatment (x) from confounders (w)
#             to get the predicted treatement (px).
#     In stage II, we
#         fit a final model (yx_model) to predict y - py from x - px.
#
#     See https://arxiv.org/pdf/1608.00060.pdf for reference.
#
#     Attributes
#     ----------
#
#     Methods
#     ----------
#     _prepare4est(data, outcome, treatment, adjustment, individual=None)
#         Prepare (fit the model) for estimating various quantities including
#         ATE, CATE, ITE, and CITE.
#     estimate(data, outcome, treatment, adjustment, quantity='ATE',
#                  condition_set=None, condition=None, individual=None)
#         Integrate estimations for various quantities into a single method.
#     estimate_ate(self, data, outcome, treatment, adjustment)
#     estimate_cate(self, data, outcome, treatment, adjustment,
#                       condition_set, condition)
#     estimate_ite(self, data, outcome, treatment, adjustment, individual)
#     estimate_cite(self, data, outcome, treatment, adjustment,
#                       condition_set, condition, individual)
#     """
#     # TODO: support more final models, e.g., non-parametric models.
#
#     def __init__(self, y_model, x_model, yx_model):
#         super().__init__()
#         if type(y_model) is str:
#             y_model = self.ml_model_dic[y_model]
#         if type(x_model) is str:
#             x_model = deepcopy(self.ml_model_dic[x_model])
#         if type(yx_model) is str:
#             yx_model = deepcopy(self.ml_model_dic[yx_model])
#
#         self.y_model = y_model
#         self.x_model = x_model
#         self.yx_model = yx_model
#
#     def _prepare4est(self, data, outcome, treatment, adjustment, individual=None):
#         self.y_model.fit(data[adjustment], data[outcome])
#         self.x_model.fit(data[adjustment], data[treatment])
#
#         py = self.y_model.predict(data[adjustment])
#         px = self.x_model.predict(data[adjustment])
#
#         self.yx_model.fit(data[treatment] - px, data[outcome] - py)
#         # TODO: support cate, now only support ate
#         result = self.yx_model.coef_
#         return result
#
#     def estimate_cate(self, data, outcome, treatment, adjustment,
#                       condition_set, condition):
#         raise NotImplementedError


class DML4CATE(BaseEstModel):
    r"""Double machine learning for estimating CATE.
    # TODO: convert the einstein notations in this section to the usual ones.
    # TODO: expand fij to higher orders of v.
    # TODO: add intercept to the final linear regression model

    (- Skip this if you are only interested in the implementation.)
    A typical double machine learning for CATE solves the following treatment
    effect estimation (note that we use the einstein notation here):
        y^i = f^i_j(v^k) x^j + g^i(v^k, w^l) + \epsilon
        x^j = h^j(v^k, w^l) + \eta
    where f^i_j(v^k) is the CATE conditional on V=v and takes the form
        f^i_j(v^k) = F^i_{j, k} \rho^k
    with \rho^k: v \to R being v^k in the simplest case. Thus we have
        y^i = F^i_{j, k} \rho^k x^j + g^i(v^k, w^l) + \epsilon.
    The coefficients F_j^i_k can be estimated from the newly-formed data
    (\rho^k x^j, y^i) with linear regression where F^i_{j, k} are just
    coefficients of every feature in {1, 2, ..., k*j}. For a simple example, if
    both y and x only have one dimention, then the CATE for an input with
    covariate (v^1, v^2, v^3) will be F_1v^1, F_2v^2, and F_3v^3. #TODO:
    However, note that letting \rho^k simply be v^k actually implicitly assumes
    that the value of v^k is small thus is a good approximation of \rho^k.


    (- Start of the implementation.)
    We implement a complicated version of the double machine learning same as
    the algorithm described in the [1]:
        1. Let k (cf_folds in our class) be an int. Form a k-fold random
        partition {..., (train_data_i, test_data_i), ...,
        (train_data_k, test_data_k)}.
        2. For each i, train y_model and x_model on train_data_i, then evaluate
        their performances in test_data_i whoes results will be saved as
        (y_hat_k, x_hat_k). All (y_hat_k, x_hat_k) will be the form
        (y_hat, x_hat).
        3. Define the differences
            y_prime = y - y_hat,
            x_prime = (x - x_hat) \otimes v.
        Then form the new dataset (y_prime, x_prime).
        4. Perform linear regression on the dataset (y_prime, x_prime) whoes
        coefficients will be saved in a vector F. The estimated CATE given V=v
        will just be
            F \dot v.
        On the other hand, the ATE can be simply estimated by taking average
        of F \dot v over the original data.

    Attributes
    ----------
    _is_fitted : bool
        True if the model is fitted ortherwise False.

    x_model : estimator
        Machine learning models for fitting x. Any such models should implement
        the fit and predict (also predict_proba if x is discrete) methods

    y_model : estimator
        Machine learning models for fitting y.

    yx_model : estimator
        Machine learning models for fitting the residual of y on residual of x.
        Currently this should be a linear regression model.

    adjustment_transformer : transformer
        Transformer for adjustment variables, by default None.

    covariate_transformer : transformer
        Transformer for covariate variables, by default None.

    is_discrete_treatment : bool

    categories : str or list

    random_state : int

    cf_fold : int
        The number of folds for performing cross fit, by default 1

    treat : float or ndarray
        In the case of single discrete treatment, treat should be an int or
        str in one of all possible treatment values which indicates the
        value of the intended treatment;
        in the case of multiple discrete treatment, treat should be a list
        or an ndarray where treat[i] indicates the value of the i-th intended
        treatment;
        in the case of continuous treatment, treat should be a float or a
        ndarray, by default None

    _v : np.array
        Covariate variables in the training set.

    _y_d : int
        Dimension of the outcome.

    _x_d : int
        Dimension of the treatment.

    ord_transformer : OrdinalEncoder
        Ordinal transformer of the discrete treament.

    oh_transformer : OneHotEncoder
        One hot encoder of the discrete treatment. Note that the total transformer
        is combined by the ord_transformer and oh_transformer. See comp_transformer
        for detail.

    label_dict : dict

    x_hat_dict : defaultdict(list)
        Cached values when fitting the treatment model.

    y_hat_dict : defaultdict(list)
        Cached values when fitting the outcome model.

    Methods
    ----------
    fit(data, outcome, treatment, adjustment, covariate)
        Fit the DML4CATE estimator model.

    estimate(data, treat, control, quantity)
        Estimate the causal effect.

    comp_transformer(x, categories='auto')
        Transform the discrete treatment into one-hot vectors.

    _cross_fit(model)
        Fit x_model and y_model in a cross fitting manner.

    _fit_first_stage(x_model, y_model, y, x, wv, folds)
        Fit the first stage of the double machine learning.

    _fit_second_stage(yx_model, y_prime, x_prime)
        Fit the second stage of the DML.

    _prepare4est(data)

    Reference
    ----------
    [1] V. Chernozhukov, et al. Double Machine Learning for Treatment and
        Causal Parameters. arXiv:1608.00060.
    """

    def __init__(
        self,
        x_model,
        y_model,
        yx_model=None,
        cf_fold=1,
        adjustment_transformer=None,
        covariate_transformer=None,
        random_state=2022,
        is_discrete_treatment=False,
        categories="auto",
    ):
        """
        Parameters
        ----------
        x_model : estimator
            Machine learning models for fitting x. Any such models should implement
            the fit and predict (also predict_proba if x is discrete) methods

        y_model : estimator
            Machine learning models for fitting y.

        yx_model : estimator, optional
            Machine learning models for fitting the residual of y on residual of x.

        cf_fold : int, optional
            The number of folds for performing cross fit, by default 1

        adjustment_transformer : transformer, optional
            Transformer for adjustment variables, by default None

        covariate_transformer : transformer, optional
            Transformer for covariate variables, by default None

        random_state : int, optional
            Random seed, by default 2022

        is_discrete_treatment : bool, optional
            If the treatment variables are discrete, set this as True, by default False

        categories : str, optional
        """
        self.cf_fold = cf_fold
        self.x_model = clone(x_model)
        self.y_model = clone(y_model)

        if yx_model is None:
            self.yx_model = LinearRegression()
        else:
            self.yx_model = yx_model

        self.adjustment_transformer = adjustment_transformer
        self.covariate_transformer = covariate_transformer

        self.x_hat_dict = defaultdict(list)
        self.y_hat_dict = defaultdict(list)

        self.x_hat_dict["is_fitted"].append(False)
        self.y_hat_dict["is_fitted"].append(False)

        super().__init__(
            random_state=random_state,
            is_discrete_treatment=is_discrete_treatment,
            categories=categories,
        )

    # TODO:could add a decorator in this place
    def fit(
        self,
        data,
        outcome,
        treatment,
        adjustment=None,
        covariate=None,
        **kwargs,
    ):
        """Fit the DML4CATE estimator model.

        Parameters
        ----------
        data : pandas.DataFrame
            The dataset used for training the model

        outcome : str or list of str, optional
            Names of the outcome variables

        treatment : str or list of str
            Names of the treatment variables

        adjustment : str or list of str, optional
            Names of the adjustment variables, by default None

        covariate : str or list of str, optional
            Names of the covariate variables, by default None

        Returns
        -------
        instance of DML4CATE
            The fitted estimator model
        """
        # must have adjustment to evaluate ATE,
        # must also have covaraite to evalueate CATE
        assert (
            adjustment is not None or covariate is not None
        ), "Need adjustment set or covariates to perform estimation."

        super().fit(
            data,
            outcome,
            treatment,
            adjustment=adjustment,
            covariate=covariate,
            **kwargs,
        )

        y, x, w, v = convert2array(data, outcome, treatment, adjustment, covariate)
        self._v = v
        self._y_d = y.shape[1]
        cfold = self.cf_fold

        if self.adjustment_transformer is not None and w is not None:
            w = self.adjustment_transformer.fit_transform(w)

        if self.covariate_transformer is not None and v is not None:
            v = self.covariate_transformer.fit_transform(v)

        if self.is_discrete_treatment:
            if self.categories == "auto" or self.categories is None:
                categories = "auto"
            else:
                categories = list(self.categories)

            # convert discrete treatment features to onehot vectors
            # note that when there are multiple treatments, we should convert
            # the problem into a single discrete treatment problem
            x = self.comp_transormer(x, categories)

        self._x_d = x.shape[1]
        wv = get_wv(w, v)

        # step 1: split the data
        if cfold > 1:
            cfold = int(cfold)
            folds = [KFold(n_splits=cfold).split(x), KFold(n_splits=cfold).split(y)]
        else:
            folds = None

        # step 2: cross fit to give the estimated y and x
        self.x_hat_dict, self.y_hat_dict = self._fit_1st_stage(
            self.x_model, self.y_model, y, x, wv, folds=folds
        )
        x_hat = self.x_hat_dict["paras"][0].reshape((x.shape))
        y_hat = self.y_hat_dict["paras"][0].reshape((y.shape))

        # step 3: calculate the differences
        x_diff = x - x_hat
        y_prime = y - y_hat
        v = np.hstack([np.ones((v.shape[0], 1)), v])
        x_prime = nd_kron(x_diff, v)

        # step 4: fit the regression problem
        self._fit_2nd_stage(self.yx_model, x_prime, y_prime)

        self._is_fitted = True

        return self

    def estimate(
        self,
        data=None,
        treat=None,
        control=None,
        quantity=None,
    ):
        """Estimate the causal effect.

        Parameters
        ----------
        data : pd.DataFrame, optional
            The test data for the estimator to evaluate the causal effect, note
            that the estimator directly evaluate all quantities in the training
            data if data is None, by default None

        treat : float or ndarray, optional
            In the case of single discrete treatment, treat should be an int or
            str of one of all possible treatment values which indicates the
            value of the intended treatment;
            in the case of multiple discrete treatment, treat should be a list
            or an ndarray where treat[i] indicates the value of the i-th intended
            treatment, for example, when there are multiple
            discrete treatments, array(['run', 'read']) means the treat value of
            the first treatment is taken as 'run' and that of the second treatment
            is taken as 'read';
            in the case of continuous treatment, treat should be a float or a
            ndarray, by default None

        control : float or ndarray, optional
            This is similar to the cases of treat, by default None

        quantity : str, optional
            The possible values of quantity include:
                'CATE' : the estimator will evaluate the CATE;
                'ATE' : the estimator will evaluate the ATE;
                None : the estimator will evaluate the ITE or CITE, by default None

        Returns
        -------
        ndarray
            The estimated causal effect with the type of the quantity.
        """
        fij = self._prepare4est(data=data)

        if hasattr(self, "treat") and treat is None:
            treat = self.treat
        if hasattr(self, "control") and control is None:
            control = self.control

        dis_tr = self.is_discrete_treatment

        if not isinstance(treat, np.ndarray) or dis_tr:
            treat = get_tr_ctrl(
                treat,
                self.comp_transormer,
                treat=True,
                one_hot=False,
                discrete_treat=dis_tr,
            )

        if not isinstance(treat, np.ndarray) or dis_tr:
            control = get_tr_ctrl(
                control,
                self.comp_transormer,
                treat=False,
                one_hot=False,
                discrete_treat=dis_tr,
            )
        self.treat = treat

        if self.is_discrete_treatment:
            effect = fij[:, :, treat] - fij[:, :, control]
        else:
            if isinstance(treat, np.ndarray):
                treat = treat.reshape(-1, self._x_d)
            if isinstance(control, np.ndarray):
                control = control.reshape(-1, self._x_d)
                effect = np.einsum("nji, ni->nji", fij, treat - control)
            else:
                effect = fij * (treat - control)

        if quantity == "CATE":
            assert self.covariate is not None
            return effect.mean(axis=0)
        elif quantity == "ATE":
            return effect.mean(axis=0)
        else:
            return effect

    def comp_transormer(self, x, categories="auto"):
        """Transform the discrete treatment into one-hot vectors.

        Parameters
        ----------
        x : ndarray, shape (n, x_d)
            An array containing the information of the treatment variables

        categories : str or list, optional
            by default 'auto'

        Returns
        -------
        ndarray
            The transformed one-hot vectors
        """
        if x.shape[1] > 1:
            if not self._is_fitted:
                self.ord_transformer = OrdinalEncoder(categories=categories)
                self.ord_transformer.fit(x)

                labels = [np.arange(len(c)) for c in self.ord_transformer.categories_]
                labels = cartesian(labels)
                categories = [np.arange(len(labels))]

                self.label_dict = {tuple(k): i for i, k in enumerate(labels)}

            x_transformed = self.ord_transformer.transform(x).astype(int)
            x = np.full((x.shape[0], 1), np.NaN)

            for i, x_i in enumerate(x_transformed):
                x[i] = self.label_dict[tuple(x_i)]

        if not self._is_fitted:
            self.oh_transformer = OneHotEncoder(categories=categories)
            self.oh_transformer.fit(x)

        x = self.oh_transformer.transform(x).toarray()

        return x

    def effect_nji(self, data=None, control=0):
        y_nji = self._prepare4est(data=data)
        n, x_d = y_nji.shape[0], y_nji.shape[2]

        if self.is_discrete_treatment:
            temp_y = y_nji[:, :, 0].reshape(n, -1, 1)
            temp_y = np.repeat(temp_y, x_d, axis=2)
            y_nji = y_nji - temp_y

        return y_nji

    def _prepare4est(self, data=None, *args, **kwargs):
        """Prepare for the estimation of causal quantities.

        Parameters
        ----------
        data : pd.DataFrame, optional
            The test data for evaluating the causal quantities, note that
            the estimator will perform the estimation in the training data if
            data is None, by default None

        Returns
        -------
        ndarray, shape (n, y_d, x_d)
            The element in the slot of (i, j, k) indicating the causal effect of
            treatment k on the j-th dimension of the outcome for the i-th example
        """
        assert (
            self.x_hat_dict["is_fitted"][0]
            and self.y_hat_dict["is_fitted"][0]
            and self._is_fitted
        ), "x_model and y_model should be"
        "trained before estimation."

        x_d, y_d = self._x_d, self._y_d
        v = self._v if data is None else convert2array(data, self.covariate)[0]

        if self.covariate_transformer is not None and v is not None:
            v = self.covariate_transformer.transform(v)
        v = np.hstack([np.ones((v.shape[0], 1)), v])
        n, v_d = v.shape[0], v.shape[1]

        # may need modification for multi-dim outcomes.
        # the reason we use transpose here is because coef f^i_{j, k}
        # (originally is a tensor, but here we treat it as a matrix
        # because we convert v^k x^j to a (k*j-dim) vector) has the
        # shape (y_d, x_d*v_d), the coefs for treatment in dim i are saved in
        # f[i*vd:(i+1)*vd]
        # TODO: rewrite this with einsum
        coef = self.yx_model.coef_
        # fij will be the desired treatment effect for treatment i on outcome j
        fij = np.full((n, y_d, x_d), np.NaN)

        def cal_fij(coef_vec, v):
            coef_matrix = coef_vec.reshape(-1, v_d)
            return coef_matrix.dot(v.reshape(-1, 1)).reshape(1, -1)

        # TODO: rewrite the following code with einsum for efficiency
        if y_d > 1:
            for j in range(y_d):
                for n, vn in enumerate(v):
                    fij[n, j, :] = cal_fij(coef[j], vn)
        else:
            coef_matrix = coef.reshape(-1, v_d)
            for n, vn in enumerate(v):
                fij[n, 0, :] = coef_matrix.dot(vn.reshape(-1, 1)).reshape(1, -1)

        return fij

    def _cross_fit(self, model, *args, **kwargs):
        folds = kwargs.pop("folds")
        is_ymodel = kwargs.pop("is_ymodel")
        target = kwargs.pop("target")
        fitted_result = defaultdict(list)

        if not is_ymodel and self.is_discrete_treatment:
            # convert back to a vector with each dimension being a value
            # indicating the corresponding discrete value
            target_converted = convert4onehot(target)
        else:
            target_converted = target

        if folds is None:
            wv = args[0]
            model.fit(wv, target_converted, **kwargs)

            if not is_ymodel and self.is_discrete_treatment:
                p_hat = model.predict_proba(wv)
            else:
                p_hat = model.predict(wv)

            fitted_result["models"].append(clone(model))
            fitted_result["paras"].append(p_hat)
            idx = np.arange(start=0, stop=wv.shape[0])
            fitted_result["train_test_id"].append((idx, idx))
        else:
            fitted_result["paras"].append(np.ones_like(target) * np.nan)

            for i, (train_id, test_id) in enumerate(folds):
                model_ = clone(model)
                temp_wv = args[0][train_id]
                temp_wv_test = args[0][test_id]
                target_train = target_converted[train_id]
                model_.fit(temp_wv, target_train, **kwargs)

                if not is_ymodel and self.is_discrete_treatment:
                    target_predict = model_.predict_proba(temp_wv_test)
                else:
                    target_predict = model_.predict(temp_wv_test)
                    # test_shape = kwargs['target'][test_id].shape
                    # if target_predict.shape != test_shape:
                    #     target_predict.reshape(test_shape)

                fitted_result["models"].append(model_)
                fitted_result["paras"][0][test_id] = target_predict
                fitted_result["train_test_id"].append((train_id, test_id))

        fitted_result["is_fitted"] = [True]

        return fitted_result

    def _fit_1st_stage(self, x_model, y_model, y, x, wv, folds=None, **kwargs):
        """Fit the models in the first stage.

        Parameters
        ----------
        x_model : estimator
            Any x_model should have the fit and predict (also predict_proba if
            x is discrete) methods.

        y_model : estimator
            Any y_model should have the fit and predict (also predict_proba if
            y is discrete) methods.

        y : ndarray, shape (n, y_d)
            The outcome vector

        x : ndarray, shape (n, x_d)
            The treatment vector

        wv : ndarray, shape (n, w_d + v_d)
            The covariate and adjustment vector

        folds : sklearn.model_selection.KFold, optional

        Returns
        -------
        tuple of dict
            The first dict containes the values of the fitted y_model and the
            second dict the fitted x_model
        """
        if self._x_d == 1:
            x = np.ravel(x)
        if self._y_d == 1:
            y = np.ravel(y)

        if folds is not None:
            x_folds, y_folds = folds
        else:
            x_folds, y_folds = None, None

        logger.info(f"_fit_1st_stage: fitting x_model {type(x_model).__name__}")
        x_hat_dict = self._cross_fit(
            x_model, wv, target=x, folds=x_folds, is_ymodel=False, **kwargs
        )

        logger.info(f"_fit_1st_stage: fitting y_model {type(y_model).__name__}")
        y_hat_dict = self._cross_fit(
            y_model, wv, target=y, folds=y_folds, is_ymodel=True, **kwargs
        )
        return (x_hat_dict, y_hat_dict)

    def _fit_2nd_stage(
        self,
        yx_model,
        x_prime,
        y_prime,
    ):
        """Fit the models in the second stage.

        Parameters
        ----------
        yx_model : estimator
            _description_

        x_prime : ndarray, shape (n, x_d)
            The residuls of the treatment vector x

        y_prime : ndarray, shape (n, y_d)
            The residuls of the outcome vector y
        """
        logger.info(f"_fit_2nd_stage: fitting yx_model {type(self.yx_model).__name__}")
        yx_model.fit(x_prime, y_prime)

    # def __repr__(self) -> str:
    #     return f'Double Machine Learning Estimator'



from IPython.display import Latex


class Prob:
    r"""
    Probability distribution, e.g., the probability expression
    \sum_{w}P(v|y)[P(w|z)P(x|y)P(u)]. We will clarify below the meanings
    of our variables.


    Attributes
    ----------
    variables : set
        The variables (v in the above example) of the probability.
    
    conditional : set
        The conditional set (y in the above example).
    
    divisor : set
        Not defined yet.
    
    marginal : set
        The sum set (w in the above example) for marginalizing the probability.
    
    product : set
        If not set(), then the probability is composed of the first probability
        object (P(v|y)) and several other probabiity objects that are all saved
        in the set product, e.g., product = {P1, P2, P3} where P1 for P(w|z),
        P2 for P(x|y), and P3 for P(u) in the above example.

    Methods
    ----------
    parse()
        Return the expression of the probability distribution.
    
    show_latex_expression()
        Show the latex expression.
    """

    def __init__(self,
                 variables=set(),
                 conditional=set(),
                 divisor=set(),
                 marginal=set(),
                 product=set()):
        """
        Parameters
        ----------
        variables : set
        
        conditional : set
        
        marginal : set
            elements are strings, summing over these elements will return the
            marginal distribution
        
        product : set
            set of Prob
        """
        self.divisor = divisor
        self.variables = variables
        self.conditional = conditional
        self.marginal = marginal
        self.product = product

    def parse(self):
        """
        Return the expression of the probability distribution.

        Returns
        ----------
        expression : str
        """
        # TODO
        expression = ''
        # First find the marginal set, -> \sum_{marginal set}
        if self.marginal:
            mar = ', '
            mar = mar.join(self.marginal)
            expression = expression + '\\sum_{' + f'{mar}' + '}'

        # Find the variables, -> \sum_{marginal} P(variables|conditional)
        if self.variables:
            var = ', '
            var = var.join(self.variables)
            if self.conditional:
                cond = ', '
                cond = cond.join(self.conditional)
                expression = expression + f'P({var}|{cond})'
            else:
                expression += f'P({var})'

        # Find the products, ->
        # \sum_{marginal} P(var|cond) \sum_{prod1.mar}P(prod1.var|prod1.cond)..
        if self.product:
            for p in self.product:
                expression = expression + '\\left[' + p.parse() + '\\right]'

        return expression

    def show_latex_expression(self):
        """
        Show the latex expression.
        """
        return Latex(f'${self.parse()}$')

    # def __repr__(self) -> str:
    #     return 'Prob'


import numpy as np
from sklearn.preprocessing import OneHotEncoder, OrdinalEncoder


from .base_models import BaseEstModel
from .utils import (convert2array, get_wv, convert4onehot,
                    get_tr_ctrl, cartesian, get_groups)


class ApproxBound(BaseEstModel):
    """
    A model used for estimating the upper and lower bounds of the causal effects.
    
    Attributes
    ----------
    y_model : MLModel, optional.
        Machine learning models for fitting the relation between treatment
        and outcome.
    
    x_prob : np.array of float, optional. Defaults to None.
        The probability of taking specific treatments.
    
    x_model : MLModel, optional. Defaults to None.
        Machine learning models for fitting the relation between treatment
        and condition set if condition set is True.
    
    random_state : int
    
    is_discrete_treatent : bool
    
    categories : str or list, optional. Deafaults to 'auto'
    
    Methods
    ----------
    fit(data, outcome, treatment, covariate=None, is_discrete_covariate=False,
        **kwargs)
    
    estimate(data=None, treat=None, control=None, y_upper=None, y_lower=None, assump=None,)
        Estimate the approximation bound of the causal effect of the treatment
        on the outcome.
    
    comp_transformer(x, categories='auto')
        Transform the discrete treatment into one-hot vectors.
    
    _prepare4est(data, treat, control, y_upper, y_lower,)
    """
    # TODO: may consider multiple treatments

    def __init__(
        self,
        y_model,
        x_prob=None,
        x_model=None,
        random_state=2022,
        is_discrete_treatment=True,
        categories='auto'
    ):
        """
        Parameters
        ----------
        y_model : estimator, optional
            Any valid y_model should implement the fit() and predict() methods
        
        x_prob : ndarray of shape (c, ), optional. Default to None
            An array of probabilities assigning to the corresponding values of x
            where c is the number of different treatment classes. All elements
            in the array are positive and sumed to 1. For example, x_prob = 
            array([0.5, 0.5]) means both x = 0 and x = 1 take probability 0.5.
            Please set this as None if you are using multiple treatments.
        
        x_model : estimator, optional. Default to None
            Models for predicting the probabilities of treatment. Any valid x_model should implement the fit() and predict_proba() methods.
        
        random_state : int, optional. Defaults to 2022.
        
        is_discrete_treatment : bool, optional
            True if the treatment is discrete.
        
        categories : str, optional
            _description_, by default 'auto'
        """
        self.y_model = y_model
        self.x_prob = x_prob
        self.x_model = x_model

        super().__init__(
            random_state,
            is_discrete_treatment,
            categories=categories
        )

    def fit(
        self,
        data,
        outcome,
        treatment,
        covariate=None,
        is_discrete_covariate=False,
        **kwargs
    ):
        """Fit x_model and y_model.

        Parameters
        ----------
        data : pandas.DataFrame
            Training data.
        
        outcome : list of str, optional
            Names of the outcome.
        
        treatment : list of str, optional
            Names of the treatment.
        
        covariate : list of str, optional. Defaults to None
            Names of the covariate.
        
        is_discrete_covariate : bool, optional. Defaults to False.

        Returns
        -------
        instance of ApproxBound
            The fitted instance of ApproxBound.

        Raises
        ------
        ValueError
            Raise error when the treatment is not discrete.
        """
        super().fit(
            data, outcome, treatment,
            covariate=covariate,
            is_discrete_covariate=is_discrete_covariate
        )

        y, x, v = convert2array(data, outcome, treatment, covariate)
        # self._y_max = y.max(axis=0)
        # self._y_min = y.min(axis=0)
        self._y = y
        self._n = len(data)

        if not self.is_discrete_treatment:
            raise ValueError(
                'The method only supports discrete treatments currently.'
            )

        if self.categories == 'auto' or self.categories is None:
            categories = 'auto'
        else:
            categories = list(self.categories)

        x_one_hot = self.comp_transormer(x, categories=categories)
        # self._num_treatments = len(self.treatment_transformer.categories_)
        self._x_d = x_one_hot.shape[1]

        self.x_label = np.array(convert4onehot(x_one_hot).astype(int))

        if self.x_prob is None:
            if v is not None:
                if self.is_discrete_covariate:
                    v = OneHotEncoder().fit_transform(v)

                self.x_model.fit(v, self.x_label)
                self.x_prob = self.x_model.predict_proba(v)
            else:
                self.x_prob = np.unique(self.x_label, return_counts=True)[1]
                self.x_prob = (self.x_prob / self._n).reshape(1, -1)
        else:
            # TODO: modify the following line for multiple treatment
            self.x_prob = self.x_prob.reshape(1, -1)

        self._v = v
        xv = get_wv(x_one_hot, v)
        self.y_model.fit(xv, y.squeeze(), **kwargs)

        self._is_fitted = True

        return self

    def estimate(
        self,
        data=None,
        treat=None,
        control=None,
        y_upper=None,
        y_lower=None,
        assump=None,
        **kwargs
    ):
        """Estimate the approximation bound of the causal effect of the treatment
        on the outcome.

        Parameters
        ----------
        data : pandas.DataFrame, optional. Default to None
            Test data. The model will use the training data if set as None.
        
        treat : ndarray of str, optional. Defaults to None
            Values of the treatment group. For example, when there are multiple
            discrete treatments, array(['run', 'read']) means the treat value of
            the first treatment is taken as 'run' and that of the second treatment
            is taken as 'read'.
        
        control : ndarray of str, optional. Defaults to None
            Values of the control group.
        
        y_upper : float. Defaults to None
            The upper bound of the outcome.
        
        y_lower : float. Defaults to None
            The lower bound of the outcome.
        
        assump : str, optional.  Defaults to 'no-assump'.
            Options for the returned bounds. Should be one of
                1. no-assump: calculate the no assumption bound whose result
                    will always contain 0.
                2. non-negative: The treatment is always positive.
                3. non-positive: The treatment is always negative.
                4. optimal: The treatment is taken if its effect is positive.


        Returns
        -------
        tuple
            The first element is the lower bound while the second element is the
            upper bound. Note that if covariate is provided, all elements are 
            ndarrays of shapes (n, ) indicating the lower and upper bounds of 
            corresponding examples where n is the number of examples. 

        Raises
        ------
        Exception
            Raise Exception if the model is not fitted.
        
        Exception
            Raise Exception if the assump is not given correctly.
        """
        if not self._is_fitted:
            raise Exception(
                'The estimator is not fitted yet. Pleas call the fit method first.'
            )

        normal, optim1, optim2 = self._prepare4est(
            data, treat, control, y_upper, y_lower
        )
        upper, lower = normal
        zero_ = np.zeros_like(lower)
        
        if assump == 'no-assump' or assump is None:
            return (lower, upper)
        elif assump == 'non-negative':
            return (zero_, upper)
        elif assump == 'non-positive':
            return (lower, zero_)
        elif assump == 'optimal':
            # TODO: need to update this for discrete treatment
            return (optim1, optim2)
        else:
            raise Exception(
                'Only support assumptions in no-assump, non-negative, and'
                f'non-positive. But was given {assump}'
            )

    def comp_transormer(self, x, categories='auto'):
        """Transform the discrete treatment into one-hot vectors properly.

        Parameters
        ----------
        x : ndarray, shape (n, x_d)
            An array containing the information of the treatment variables
        
        categories : str or list, optional
            by default 'auto'

        Returns
        -------
        ndarray
            The transformed one-hot vectors
        """
        if x.shape[1] > 1:
            if not self._is_fitted:
                self.ord_transformer = OrdinalEncoder(categories=categories)
                self.ord_transformer.fit(x)

                labels = [
                    np.arange(len(c)) for c in self.ord_transformer.categories_
                ]
                labels = cartesian(labels)
                categories = [np.arange(len(labels))]

                self.label_dict = {tuple(k): i for i, k in enumerate(labels)}

            x_transformed = self.ord_transformer.transform(x).astype(int)
            x = np.full((x.shape[0], 1), np.NaN)

            for i, x_i in enumerate(x_transformed):
                x[i] = self.label_dict[tuple(x_i)]

        if not self._is_fitted:
            self.oh_transformer = OneHotEncoder(categories=categories)
            self.oh_transformer.fit(x)

        x = self.oh_transformer.transform(x).toarray()

        return x

    def _prepare4est(
        self,
        data,
        treat,
        control,
        y_upper,
        y_lower,
    ):
        # x_d = self._x_d
        x_prob = self.x_prob

        treat = get_tr_ctrl(
            treat,
            self.comp_transormer,
            treat=True,
            one_hot=False,
            discrete_treat=True
        )
        control = get_tr_ctrl(
            control,
            self.comp_transormer,
            treat=False,
        )

        if self.covariate is None:
            n = 1
            v = None
            if data is None:
                y = self._y
                x = self.x_label
                # y_upper = y.max(axis=0) if y_upper is None else y_upper
                # y_lower = y.min(axis=0) if y_lower is None else y_lower
            else:
                y, x = convert2array(data, self.outcome, self.treatment)
                x = convert4onehot(self.comp_transormer(x,))
        else:
            if data is None:
                v = self._v
                y = self._y
                x = self.x_label
                n = self._n
            else:
                y, v, x = convert2array(
                    data, self.outcome, self.covariate, self.treatment
                )
                n = y.shape[0]
                x_prob = self.x_model.predict_proba(v)

        y_treat = get_groups(treat, x.reshape(-1, 1), False, y)[0]
        y_ctrl = get_groups(control, x.reshape(-1, 1), False, y)[0]
        y_tr_max, y_tr_min = y_treat.max(axis=0), y_treat.min(axis=0)
        y_ctrl_max, y_ctrl_min = y_ctrl.max(axis=0), y_ctrl.min(axis=0)

        y_upper = max(y_tr_max, y_ctrl_max) if y_upper is None else y_upper
        y_lower = min(y_tr_min, y_ctrl_min) if y_lower is None else y_lower

        # TODO: modify the following line for multiple treatment
        xt, x0 = np.zeros((n, self._x_d)), np.zeros((n, self._x_d))
        xt[:, treat] = 1
        x0[:, control] = 1

        xt = get_wv(xt, v)
        x0 = get_wv(x0, v)
        yt = self.y_model.predict(xt)
        y0 = self.y_model.predict(x0)
        xt_prob = x_prob[:, treat]
        x0_prob = x_prob[:, control]

        upper = xt_prob * yt + (1 - xt_prob) * y_upper \
            - (1 - x0_prob) * y_lower - x0_prob * y0
        lower = xt_prob * yt + (1 - xt_prob) * y_lower \
            - (1 - x0_prob) * y_upper - x0_prob * y0

        optimal_upper1 = xt_prob * yt - (1 - x0_prob) * y_lower \
            + (1 - xt_prob - x0_prob) * y_upper
        optimal_lower1 = x0_prob * lower - x0_prob * y0
        optimal_upper2 = yt - xt_prob * lower - x0_prob * y0
        optimal_lower2 = xt_prob * yt + x0_prob * lower - y0

        return ((upper, lower), (optimal_upper1, optimal_lower1),
                (optimal_upper2, optimal_lower2))



from copy import deepcopy

import numpy as np

from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import OneHotEncoder, OrdinalEncoder, PolynomialFeatures

from .base_models import BaseEstModel
from .utils import convert2array, nd_kron, get_wv, cartesian

# TODO: double check the case where is_discrete_treatment=True

# class TwoSLS(BaseEstModel):
#     # TODO: simply import the 2SLS from StatsModel can finish this
#     def __init__(
#         self,
#         random_state=2022,
#         is_discrete_treatment=False,
#         is_discrete_outcome=False,
#         categories='auto'
#     ):
#         super().__init__(
#             random_state,
#             is_discrete_treatment,
#             is_discrete_outcome,
#             categories
#         )

#     def fit(self, data, outcome, treatment, **kwargs):
#         return super().fit(data, outcome, treatment, **kwargs)

#     def estimate(self, data=None, **kwargs):
#         return super().estimate(data, **kwargs)

#     def _prepare4est(self):
#         pass


class NP2SLS(BaseEstModel):
    r"""
    See Instrumental variable estimation of nonparametric models
    (https://eml.berkeley.edu/~powell/npiv.pdf) for reference.

    This method is similar to the conventional 2SLS and is also composed of
    2 stages after finding new features of x, w, and z,
        f^d = f^d(z)
        g^\mu = g^\mu(v),
    which are represented by some non-linear functions (basis functions). These
    stages are:
        1. Fit the treatment model:
            x_hat(z, v, w) = a^{d, \mu} f_d(z)g_{\mu}(v) + h(v, w) + \eta
        2. Generate new treatments x_hat, and then fit the outcome model
            y(x_hat, v, w) = a^{m, \mu} \psi_m(x_hat)g_{\mu}(v) + k(v, w)
            + \epsilon.
    The final causal effect is then estimated as
        y(x_hat_1, v, w) - y(x_hat_0, v, w).
    """

    def __init__(
        self,
        x_model=None,
        y_model=None,
        random_state=2022,
        is_discrete_treatment=False,
        is_discrete_outcome=False,
        categories="auto",
    ):
        """

        Parameters
        ----------
        x_model : estimator, optional
            The machine learning model to model the treatment. Any valid x_model should implement the `fit` and `predict` methods, by default None
        y_model : estimator, optional
            The machine learning model to model the outcome. Any valid y_model should implement the `fit` and `predict` methods, by default None
        random_state : int, optional
            by default 2022
        is_discrete_treatment : bool, optional
            Avoid setting this as True in the current version, by default False
        is_discrete_outcome : bool, optional
            Avoid setting this as True in the current version, by default False
        categories : str, optional
            by default 'auto'
        """
        self.x_model = LinearRegression() if x_model is None else x_model
        self.y_model = LinearRegression() if y_model is None else y_model

        super().__init__(
            random_state, is_discrete_treatment, is_discrete_outcome, categories
        )

    def fit(
        self,
        data,
        outcome,
        treatment,
        instrument,
        is_discrete_instrument=False,
        treatment_basis=("Poly", 2),
        instrument_basis=("Poly", 2),
        covar_basis=("Poly", 2),
        adjustment=None,
        covariate=None,
        **kwargs,
    ):
        """Fit a NP2SLS. Note that when both treatment_basis and instrument_basis have degree
        1 we are actually doing 2SLS.

        data : DataFrame
            Training data for the model.

        outcome : str or list of str
            Names of the outcomes.

        treatment : str or list of str
            Names of the treatment vectors.

        covariate : str of list of str
            Names of the covariate vectors.

        treatment_basis : tuple of 2 elements, optional, default=('Poly', 2)
            Option for transforming the original treatment vectors. The first element indicates the transformation basis function while the second one denotes the degree. Currently only support 'Poly' in the first element.

        instrument_basis : tuple of 2 elements, optional, default=('Poly', 2)
            Option for transforming the original instrument vectors. The first element indicates the transformation basis function while the second one denotes the degree. Currently only support 'Poly' in the first element.

        covar_basis : tuple of 2 elements, optional, default=('Poly', 2)
            Option for transforming the original covariate vectors. The first element indicates the transformation basis function while the second one denotes the degree. Currently only support 'Poly' in the first element.

        is_discrete_instrument : bool, default=False
        """
        assert instrument is not None, "Instrument is required."

        if all(
            (
                treatment_basis is None,
                instrument_basis is None,
                covar_basis is None,
            )
        ):
            raise ValueError(
                "Please specifiy the non-linear transformers for variables, "
                "or use TwoSLS method instead."
            )

        super().fit(
            data,
            outcome,
            treatment,
            covariate=covariate,
            adjustment=adjustment,
            instrument=instrument,
        )
        self.is_discrete_instrument = is_discrete_instrument

        self._n = len(data)

        # data preprocessing
        y, x, z, w, v = convert2array(
            data, outcome, treatment, instrument, adjustment, covariate
        )
        self._y_d = y.shape[1]

        if self.is_discrete_treatment:
            self.treatment_transformer = OrdinalEncoder()
            self.treatment_transformer.fit(x)
            x = self.treatment_transformer.transform(x)

        if self.is_discrete_instrument:
            self.instrument_transformer = OneHotEncoder()
            self.instrument_transformer.fit(z)
            z = self.instrument_transformer.transform(z)

        if isinstance(treatment_basis, list) or isinstance(treatment_basis, tuple):
            method, degree = treatment_basis[0], treatment_basis[1]
            self._x_basis_func = self.get_basis_func(degree, method, x, **kwargs)
        else:
            self._x_basis_func = deepcopy(treatment_basis)

        if isinstance(instrument_basis, list) or isinstance(instrument_basis, tuple):
            method, degree = instrument_basis[0], instrument_basis[1]
            self._z_basis_func = self.get_basis_func(degree, method, z, **kwargs)
        else:
            self._z_basis_func = deepcopy(instrument_basis)

        if isinstance(covar_basis, str) or isinstance(covar_basis, tuple):
            method, degree = covar_basis[0], covar_basis[1]
            self._v_basis_func = self.get_basis_func(degree, method, v, **kwargs)
        else:
            self._v_basis_func = deepcopy(covar_basis)

        self._x_d = x.shape[1]

        # get expansion basis functions for parameters
        x_transformed = self._x_basis_func.fit_transform(x)
        z_transformed = self._z_basis_func.fit_transform(z)
        v_transformed = self._v_basis_func.fit_transform(v)

        self._v_transformed = v_transformed
        self._w = w

        if v_transformed is not None:
            zv = nd_kron(z_transformed, v_transformed)
        else:
            zv = z_transformed

        zvw = get_wv(zv, w, np.ones((zv.shape[0], 1)))

        # stage 1: fit the x_model on z and w, v
        self.x_model.fit(zvw, x_transformed)

        # stage 2: fit the y_model on x, and w, v
        x_hat = self.x_model.predict(zvw)

        if v_transformed is not None:
            xv = nd_kron(x_hat, v_transformed)
        else:
            xv = x_hat

        xvw = get_wv(xv, w, np.ones((xv.shape[0], 1)))
        self.y_model.fit(xvw, y)

        # set _is_fitted as True
        self._is_fitted = True

        return self

    def estimate(
        self,
        data=None,
        treat=None,
        control=None,
        quantity=None,
        marginal_effect=False,
    ):
        if not self._is_fitted:
            raise Exception("The estimator is not fitted yet.")

        if hasattr(self, "treat") and treat is None:
            treat = self.treat
        if hasattr(self, "control") and control is None:
            control = self.control

        yt, y0 = self._prepare4est(
            data=data,
            treat=treat,
            control=control,
            marginal_effect=marginal_effect,
            quantity=quantity,
        )
        if quantity == "CATE":
            assert (
                self.covariate is not None
            ), "Need caovariate to compute the CATE in this case."
            return (yt - y0).mean(dim=0)
        if quantity == "ATE":
            return (yt - y0).mean(dim=0)
        elif quantity is None:
            return yt - y0
        else:
            assert (
                quantity == "CF"
            ), f'quantity should be one of "ATE", "CATE", "CF", and None, but was given {quantity}'
            return yt  # counterfactuals

    def effect_nji(self, data=None):
        if self.is_discrete_treatment:
            v, w, x, n = self._check_data(data=data)

            if self._x_d == 1:
                x_d = len(self.treatment_transformer.categories_[0])
                y_nji = np.full((n, self._y_d, x_d), np.nan)

                for treat in range(x_d):
                    xt = np.repeat(np.array([[treat]]), n, axis=0)
                    # xt = self._x_basis_func.fit_transform(xt)
                    xt = self._x_basis_func.transform(xt)
                    xv_t = nd_kron(xt, v) if v is not None else xt
                    xvw_t = get_wv(xv_t, w, np.ones((n, 1)))
                    y_pred = self.y_model.predict(xvw_t).reshape(-1, self._y_d)

                    y_nji[:, :, treat] = y_pred

                y_ctrl = y_nji[:, :, 0].reshape(n, -1, 1).repeat(x_d, aixs=2)
            else:
                labels = [
                    np.arange(len(i)) for i in self.treatment_transformer.categories_
                ]
                _treatments = cartesian(labels)

                y_nji = np.full((n, self._y_d, len(_treatments)), np.nan)
                for i, treat in enumerate(_treatments):
                    xt = np.repeat(treat.reshape(1, -1), n, axis=0)
                    xt = self._x_basis_func.transform(xt)
                    xv_t = nd_kron(xt, v) if v is not None else xt
                    xvw_t = get_wv(xv_t, w, np.ones((n, 1)))
                    y_pred = self.y_model.predict(xvw_t).reshape(-1, self._y_d)

                    y_nji[:, :, i] = y_pred

                y_ctrl = (
                    y_nji[:, :, 0].reshape(n, -1, 1).repeat(len(_treatments), axis=2)
                )

        else:
            yt, y0 = self._prepare4est(data=data, marginal_effect=False)
            # TODO: the definition of effect_nji may need to be modified in the future verison in this case
            if yt.ndim == 1:
                yt = yt.reshape(-1, 1)
                y0 = y0.reshape(-1, 1)

            n, y_d = yt.shape
            y_nji = np.full(n, y_d, 2)

            y_nji[:, :, 0] = y0
            y_nji[:, :, 1] = yt
            y_ctrl = y0.reshape(n, -1, 1).repeat(2, axis=2)

        y_nji = y_nji - y_ctrl

        return y_nji

    def _prepare4est(
        self,
        data,
        treat,
        control,
        marginal_effect,
        quantity=None,
    ):
        v, w, x, n = self._check_data(data=data)

        if quantity != "CF":
            if treat is not None:
                if not isinstance(treat, np.ndarray):
                    treat = np.array([treat]).reshape(1, -1)
                # assert (
                #     treat.shape[1] == self._x_d
                # ), f"Expect dimension of treat as {self._x_d}, but was given {treat.shape[1]}"
                treat = treat.reshape(-1, self._x_d)
            else:
                treat = np.array([[1 for i in range(self._x_d)]])

            if treat.shape[0] == 1:
                xt = np.repeat(treat, n, axis=0)
            else:
                xt = treat
                assert (
                    xt.shape[0] == n
                ), f"The number of treat {xt.shape[0]} does not match the number of the samples"

            if control is not None:
                if not isinstance(control, np.ndarray):
                    control = np.array([control]).reshape(1, -1)
                control = control.reshape(-1, self._x_d)
                # assert (
                #     control.shape[1] == self._x_d
                # ), f"Expect dimension of control as {self._x_d}, but was given {control.shape[1]}"
            else:
                control = np.array([[0 for i in range(self._x_d)]])

            if control.shape[0] == 1:
                x0 = np.repeat(control, n, axis=0)
            else:
                x0 = control
                assert (
                    x0.shape[0] == n
                ), f"The number of control {x0.shape[0]} does not match the number of the samples"

            if self.is_discrete_treatment:
                xt = self.treatment_transformer.transform(xt)
                x0 = self.treatment_transformer.transform(x0)
            # else:
            #     if treat is not None and not isinstance(treat, np.ndarray):
            #         treat = np.array([treat]).reshape(-1, 1)
            #     else:
            #         treat = np.repeat(np.array([[1 for i in range(self._x_d)]]), n, axis=0)
            #     treat = 1 if treat is None else treat
            #     control = 0 if control is None else control

            # if not isinstance(treat, np.ndarray):
            #     xt = np.repeat(np.array([[treat for i in range(self._x_d)]]), n, axis=0)
            #     x0 = np.repeat(
            #         np.array([[control for i in range(self._x_d)]]), n, axis=0
            #     )
            # else:
            #     xt = treat
            #     x0 = control
            # xt = np.repeat(np.array([[treat]]), n, axis=0)
            # x0 = np.repeat(np.array([[control]]), n, axis=0)
        else:
            xt = x
            x0 = deepcopy(xt)

        # xt = self._x_basis_func.fit_transform(xt)
        # x0 = self._x_basis_func.fit_transform(x0)
        xt = self._x_basis_func.transform(xt)
        x0 = self._x_basis_func.transform(x0)

        if v is not None:
            xv_t = nd_kron(xt, v)
            xv_0 = nd_kron(x0, v)
        else:
            xv_t = xt
            xv_0 = x0

        xvw_t = get_wv(xv_t, w, np.ones((n, 1)))
        xvw_0 = get_wv(xv_0, w, np.ones((n, 1)))

        yt = self.y_model.predict(xvw_t)
        y0 = self.y_model.predict(xvw_0)

        return yt, y0

    def _check_data(self, data):
        if data is None:
            v = self._v_transformed
            w = self._w
            x = None
            n = self._n
        else:
            w, x, v = convert2array(
                data, self.adjustment, self.treatment, self.covariate
            )
            v = self._v_basis_func.transform(v) if v is not None else None

            n = len(data)

        return v, w, x, n

    def get_basis_func(self, degree, func, para, **kwargs):
        if func == "Poly":
            return self._poly_basis_func(degree, para, **kwargs)
        elif func == "Hermite":
            return self._hermite_basis_func(degree, para, **kwargs)
        else:
            raise ValueError(f"Do not support {func} currently.")

    def _poly_basis_func(self, degree, para, **kwargs):
        poly = PolynomialFeatures(degree=degree, **kwargs)
        return poly

    def _hermite_basis_func(self, degree, para, **kwargs):
        raise NotImplementedError



import numpy as np
import torch
import torch.nn as nn

from sklearn import linear_model

from .base_models import BaseEstModel
# TODO: consider treatments other than binary treatment.


class PropensityScore:
    """
    The class wrapping a machine learning model to be able to predict the
    propensity score
        ps_{x=xt}(w) = P(x=xt|w)

    Attributes
    ----------
    ml_model_dic : dict
        Default model dictionary where keys are name of machine learning
        models while values are corresponding models.
    ml_model : MLModel, optional

    Methods
    ----------
    fit(data, treatment, adjustment)
    predict(data, adjustment)
    predict_proba(data, adjustment)
        Use this method when the treatment is discrete rather than binary.
    fit_predict(train_data, treatment, adjustment, pre_data)
        Combination of fit() and predict(), where the model is trained in the
        train_data and predict the pre_data if it's not None, otherwise the
        model will predict labels of the train_data.
    """

    def __init__(self, ml_model=None):
        """
        Parameters
        ----------
        ml_model : str, optional. Defaults to None.
        """
        self.ml_model_dic = {
            'LogisticR': linear_model.LogisticRegression()
        }

        if ml_model is None:
            ml_model = self.ml_model_dic['LogisticR']
        if type(ml_model) is str:
            ml_model = self.ml_model_dic[ml_model]

        self.ml_model = ml_model

    def fit(self, data, treatment, adjustment):
        self.ml_model.fit(data[adjustment], data[treatment])

    def predict(self, data, adjustment):
        return self.ml_model.predict(data[adjustment])

    def predict_proba(self, data, adjustment, target=None):
        p = self.ml_model.predict_proba(data[adjustment])
        try:
            p_ = p.transpose()[target]
        except TypeError:
            p_ = p.transpose(0, 1)[target]
        return p_

    def fit_predict(self, train_data, treatment, adjustment, pre_data=None):
        """Combination of fit() and predict(), where the model is trained in the
        train_data and predict the pre_data if it's not None, otherwise the
        model will predict labels of the train_data.

        Parameters
        ----------
        train_data : pd.DataFrame
        treatment : str
        adjustment : list of str
        pre_data : pd.DataFrame. Defaults to None

        Returns
        ----------
        np.array
        """
        self.ml_model.fit(train_data, treatment, adjustment)
        if pre_data is None:
            pre_data = train_data
        return self.predict(pre_data, adjustment)


class InversePbWeighting(BaseEstModel):
    r"""
    Inverse Probability Weighting. The identification equation is defined as
        E[y|do(x)] = E[I(X=x)y / P(x|W)],
    where I is the indicator function and W is the adjustment set.
    For binary treatment, we have
        ATE = E[y|do(x=1) - y|do(x=0)] =
            E[I(X=1)y / e(W)] - E[E[I(X=0)y / (1 - ps(w))]
    where ps(w) is the propensity score
        ps(w) = P(x|W).
    Therefore, the final estimated ATE should be
        1 / n_1 \sum_{i| x_i = 1} y_i / e(w_i)
            - 1 / n_2 \sum_{j| x_j = 0} y_j / (1 - e(w_i)).

    Attributes
    ----------
    ml_model_dic : dict
        A dictionary of default machine learning sklearn models currently
        including
            'LR': LinearRegression
            'LogisticR': LogisticRegression.

    ps_model : PropensityScore
        The wrapped machine learning model for the propensity score.

    Methods
    ----------
    _prepare4est(data, outcome, treatment, adjustment, individual=None)
        Prepare (fit the model) for estimating various quantities including
        ATE, CATE, ITE, and CITE.
    estimate(data, outcome, treatment, adjustment, quantity='ATE',
                 condition_set=None, condition=None, individual=None)
        Integrate estimations for various quantities into a single method.
    estimate_ate(self, data, outcome, treatment, adjustment)
    estimate_cate(self, data, outcome, treatment, adjustment,
                      condition_set, condition)
    estimate_ite(self, data, outcome, treatment, adjustment, individual)
    estimate_cite(self, data, outcome, treatment, adjustment,
                      condition_set, condition, individual)

    """
    # TODO: support more methods, treatments other than binary treatment.

    def __init__(self, ps_model):
        super().__init__()
        if type(ps_model) is str:
            ps_model = self.ml_model_dic[ps_model]

        self.ps_model = PropensityScore(ml_model=ps_model)

    def _prepare4est(self, data, outcome, treatment, adjustment, individual=None):
        self.ps_model.fit(data, treatment, adjustment)
        ps = self.ps_model.predict_proba(data, adjustment, target=1)
        o = np.ones(len(ps))
        eps = 1e-7  # numerical stability
        # The following computation is valid only for binary treatment.
        # TODO: consider continuous treatment.
        result = (data[treatment] * data[outcome] + eps) / (ps + eps) \
            - ((o - data[treatment]) * data[outcome] + eps) / (o - ps + eps)
        return result

    # The following method is the old version.
    # def estimate_ate(self, data, outcome, treatment, adjustment):
    #     self.ps_model.fit(data, treatment, adjustment)
    #     t1_data = data.loc[data[treatment] > 0]
    #     t0_data = data.loc[data[treatment] <= 0]
    #     t1_ew = self.ps_model.predict(t1_data, adjustment)
    #     t0_ew = np.ones(len(t0_data)) \
    #         - self.ps_model.predict(t0_data, adjustment)
    #     result = (t1_data[outcome] / t1_ew).mean() \
    #         - (t0_data[outcome] / t0_ew).mean()
    #     return result



# from . import graph
# from . import model
# from . import prob
# from . import scm
# #from . import identification
from .graph import CausalGraph
from .model import CausalModel
from .prob import Prob
from .scm import CausalStructuralModel



from re import I
import networkx as nx

from copy import deepcopy
from itertools import combinations


def ancestors_of_iter(g, x):
    """Return the ancestors of all nodes in x.

    Parameters
    ----------
    x : set of str
        a set of nodes in the graph

    Returns
    ----------
    set of str
        Ancestors of nodes x of the graph
    """
    an = set()
    x = {x} if isinstance(x, str) else x
    
    for node in x:
        an.add(node)
        try:
            an.update(nx.ancestors(g, node))
        except Exception:
            pass

    return an    


def descendents_of_iter(g, x):
    """Return the descendents of all nodes in x.

    Parameters
    ----------
    x : set of str
        a set of nodes in the graph

    Returns
    ----------
    set of str
        Descendents of nodes x of the graph
    """
    des = set()
    x = {x} if isinstance(x, str) else x
    
    for node in x:
        des.add(node)
        try:
            des.update(nx.descendants(g, node))
        except Exception:
            pass

    return des

def remove_ingo_edges(graph, incoming=True, *S):
    """Remove incoming or outgoing edges for nodes of S in the graph.

    Parameters
    ----------
    graph : nx.DiGraph, optional
        
    incoming : bool, optional
        If True, then all incoming edges of nodes in S will be removed,
        otherwise remove all outgoing edges of nodes in S, by default True

    Returns
    -------
    nx.DiGraph, optional
        A deepcopy of graph where the incoming edges or outgoing edges of nodes
        in S are removed.
    """
    g = deepcopy(graph)
    S = filter(None, S)
    
    for i in S:
        i = {i} if isinstance(i, str) else i
        check_nodes(g.nodes, i)
        for node in i:
            if incoming:
                g.remove_edges_from([(p, node) for p in g.predecessors(node)])
            else:
                g.remove_edges_from([(node, c) for c in g.successors(node)])

    return g

def check_nodes(nodes=None, *S):
    """Check if nodes contained in S are present in nodes.

    Parameters
    ----------
    nodes : list, optional
        A list of nodes which should include all nodes in S, by default None
    """
    S = filter(None, S)
    for i in S:
        i = {i} if isinstance(i, str) else i
        for j in i:
            assert j in nodes, f'The node {j} is not in all avaliable nodes {nodes}'


def check_ancestors_chain(dag, node, U):
    an = nx.ancestors(dag, node)

    if len(an) == 0:
        return True
    elif U in an:
        return False
    else:
        for n in an:
            if not check_ancestors_chain(dag, n, U):
                return False

        return True


def powerset(iterable):
    """Return the power set of the iterable.

    Parameters
    ----------
    iterable : container
        Can be a set or a list.

    Returns
    ----------
    list
        The list of power set.
    """
    # TODO: return a generator instead of a list
    s = list(iterable)
    power_set = []
    for i in range(len(s) + 1):
        for j in combinations(s, i):
            power_set.append(set(j))
    return power_set


class IdentificationError(Exception):
    def __init__(self, info):
        super().__init__(self)
        self.info = info

    def __str__(self):
        return self.info



from collections import OrderedDict

import numpy as np
import pandas as pd

from ylearn.utils import to_repr


class BaseDiscovery:
    def __call__(self, data, *, return_dict=False, threshold=None, **kwargs):
        raise NotImplementedError()

    # @staticmethod
    # def _matrix2dict_(matrix, names=None, threshold=0.1):
    #     assert isinstance(matrix, np.ndarray)
    #     assert matrix.ndim == 2 and matrix.shape[0] == matrix.shape[1]
    #     assert feature_names is None or len(feature_names) == matrix.shape[0]
    #
    #     if names is None:
    #         names = [f'X{i}' for i in range(matrix.shape[0])]
    #
    #     df = pd.DataFrame(matrix, columns=names, index=names)
    #     df = df.T.abs()
    #
    #     m = OrderedDict()
    #     for f in names:
    #         m[f] = df[df[f] > threshold].index.tolist()
    #
    #     return m

    @staticmethod
    def matrix2dict(matrix, threshold=0.1, names=None, ):
        assert isinstance(matrix, (np.ndarray, pd.DataFrame))
        assert matrix.ndim == 2 and matrix.shape[0] == matrix.shape[1]
        assert names is None or len(names) == matrix.shape[0]

        matrix = matrix.copy()

        if names is None and isinstance(matrix, pd.DataFrame):
            names = matrix.columns.tolist()
        if isinstance(matrix, pd.DataFrame):
            matrix = matrix.values

        for i in range(matrix.shape[0]):
            for c in range(i + 1, matrix.shape[0]):
                if abs(matrix[i, c]) < abs(matrix[c, i]):
                    matrix[i, c] = 0.
                else:
                    matrix[c, i] = 0.
            matrix[i, i] = 0.

        d = OrderedDict()
        idx = np.arange(matrix.shape[0])
        for i in range(matrix.shape[0]):
            row = matrix[i]
            to = np.where(np.where(np.abs(row) > threshold, idx, -1) >= 0)[0]
            if names is None:
                d[i] = to.tolist()
            else:
                d[names[i]] = [names[t] for t in to]

        return d

    @staticmethod
    def matrix2array(matrix, names=None, ):
        """
        Returns
        -------
            table of:
                from,to,prob
                ...
        """
        assert isinstance(matrix, (np.ndarray, pd.DataFrame))
        assert matrix.ndim == 2 and matrix.shape[0] == matrix.shape[1]
        assert names is None or len(names) == matrix.shape[0]

        if names is None and isinstance(matrix, pd.DataFrame):
            names = matrix.columns.tolist()

        r = []
        for row in range(matrix.shape[0]):
            for col in range(row + 1, matrix.shape[1]):
                if hasattr(matrix, 'iloc'):
                    v = matrix.iloc[row, col]
                    vt = matrix.iloc[col, row]
                else:
                    v = matrix[row, col]
                    vt = matrix[row, col]
                if abs(v) >= abs(vt):
                    r.append([row, col, v])
                else:
                    r.append([col, row, vt])

        if names is not None:
            r = [[names[row[0]], names[row[1]], row[2]] for row in r]

        return r

    @staticmethod
    def matrix2df(matrix, names=None, ):
        r = BaseDiscovery.matrix2array(matrix, names=names)
        return pd.DataFrame(r, columns=['from', 'to', 'prob'])

    def __repr__(self):
        return to_repr(self)



import torch
import numpy as np


def func(dat, pho=10, alpha=1, lambda1=1):
    cuda = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    x = torch.tensor(dat, device=cuda).float()
    n = x.size()[0]
    batch_num = x.size()[1]
    w = torch.rand(n, n, device=cuda, requires_grad=True)

    def h(w):
        return torch.trace(torch.matrix_exp(w * w)) / n - 1

    def linear_loss(w):
        res = x - torch.matmul(w, x)
        return torch.sum(res * res) / (n * batch_num)

    def tot_loss(w, pho, alpha):
        return linear_loss(w) + h(w) * h(w) * 0.5 * pho + alpha * h(w) + lambda1 * torch.sum(w * w) / (n * n)

    optimizer = torch.optim.SGD([w], lr=0.1, momentum=0.9)

    def local_minimize(pho, alpha):
        for i in range(200):
            optimizer.zero_grad()
            l = tot_loss(w, pho, alpha)
            l.backward()
            optimizer.step()
            # print(i,w)

    h_ = h(w.clone().detach())
    for _ in range(100):
        local_minimize(pho, alpha)
        alpha = alpha + pho * h(w.clone().detach())
        print(w)

    return w.detach().cpu().numpy()


def reg(x, y):
    cuda = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    x = torch.tensor(x, device=cuda).float()
    y = torch.tensor(y, device=cuda).float()
    size = x.size()[0]
    batch_num = x.size()[1]
    w = torch.ones([size], device=cuda, requires_grad=True).float()
    wt = torch.tensor([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], device=cuda).float()
    optimizer = torch.optim.SGD([w], lr=0.01, momentum=0.9)

    def l(w, x, y):
        res = y - torch.matmul(w, x)
        loss = torch.sum(res * res) / (size * batch_num)
        return loss

    for i in range(10000):
        optimizer.zero_grad()
        loss = l(w, x, y)
        loss.backward()
        optimizer.step()
        # w=torch.tensor(w-0.1*w.grad,device=cuda,requires_grad=True).float()

        # w = torch.tensor(w - 0.5 * w.grad,requires_grad=True).float()
        # print(w)
        print(loss)
    print(w)


def dat_gen():
    x1 = np.random.random_sample(20000)
    x2 = x1 + 0.001 * np.random.random_sample(20000)
    x3 = x2 - 0.001 * np.random.random_sample(20000)
    return np.array([x2, x1, x3])


# def dat_gen1(n, coef12=1, coef32=2, eps=1e-3, change_order=False):
#     x2 = np.random.normal(size=(n, ))
#     x1 = eps * np.random.normal(size=(n, )) + coef12 * x2
#     x3 = eps * np.random.normal(size=(n, )) + coef32 * x2
#     if change_order:
#         return np.array([x1, x2, x3])
#     else:
#         return np.array([x2, x1, x3])


if __name__ == 'main':
    dat = dat_gen()
    print(f'chain result {func(dat)}')

# dat = dat_gen1()
# print(f'folk result {func(dat)}')

# dat = dat_gen1()
# print(f'folk new order {func(dat)}')


import numbers
from math import ceil
from copy import deepcopy

import numpy as np
import pandas

# from ctypes import memset, sizeof, c_double, memmove

import sklearn
from sklearn.utils import check_random_state
from sklearn.utils import check_scalar

from sklearn.preprocessing import OrdinalEncoder

from ylearn.sklearn_ex.cloned.tree._splitter import Splitter
from ylearn.sklearn_ex.cloned.tree._splitter import BestSplitter
from ylearn.sklearn_ex.cloned.tree._splitter import RandomSplitter
from ylearn.sklearn_ex.cloned.tree._tree import DepthFirstTreeBuilder
from ylearn.sklearn_ex.cloned.tree._tree import BestFirstTreeBuilder
from ylearn.sklearn_ex.cloned.tree._tree import Tree
from sklearn.tree import plot_tree

from ..utils import logging
from .utils import (convert2array, get_wv, get_treat_control)
from .base_models import BaseEstModel

from ylearn.estimator_model._tree.tree_criterion import CMSE, MSE, HonestCMSE
# import pyximport
# pyximport.install(setup_args={"script_args": ["--verbose"]})

logger = logging.get_logger(__name__)

SPLITTERS = {"best": BestSplitter, "random": RandomSplitter}

EPS = 1e-5


class CausalTree(BaseEstModel):
    # TODO: add support for multi-output causal tree
    """A class for estimating causal effect with decision tree.
    
    Parameters
    ----------
    splitter : {"best", "random"}, default="best" 
    
        The strategy used to choose the split at each node. Supported 
        strategies are "best" to choose the best split and "random" to choose 
        the best random split.

    max_depth : int, default=None 
        The maximum depth of the tree. If None, then nodes are expanded until 
        all leaves are pure or until all leaves contain less than 
        min_samples_split samples. 

    min_samples_split : int or float, default=2 
        The minimum number of samples required to split an internal node:
        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
        `ceil(min_samples_split * n_samples)` are the minimum
        number of samples for each split.

    min_samples_leaf : int or float, default=1 
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.
        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
        `ceil(min_samples_leaf * n_samples)` are the minimum
        number of samples for each node.

    max_features : int, float or {"sqrt", "log2"}, default=None 
        The number of features to consider when looking for the best split:
            
            1. If int, then consider `max_features` features at each split.
            2. If float, then `max_features` is a fraction and `int(max_features * n_features)` features are considered at each split.
            3. If "sqrt", then `max_features=sqrt(n_features)`.
            4. If "log2", then `max_features=log2(n_features)`.
            5. If None, then `max_features=n_features`.

    random_state : int 
        Controls the randomness of the estimator.

    max_leaf_nodes : int, default to None 
        Grow a tree with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    min_impurity_decrease : float, default=0.0 
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.
        The weighted impurity decrease equation is the following::
            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)
        where ``N`` is the total number of samples, ``N_t`` is the number of
        samples at the current node, ``N_t_L`` is the number of samples in the
        left child, and ``N_t_R`` is the number of samples in the right child.
        ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
        if ``sample_weight`` is passed.

    ccp_alpha : non-negative float, default to 0.0 
        Value for pruning the tree. *Not implemented yet*.
    
    categories : str, optional. Defaults to 'auto'. 

    See Also
    --------
    BaseDecisionTree : The default implementation of decision tree in sklearn.

    Methods
    ----------
    fit(data, outcome, treatment, adjustment=None, covariate=None, treat=None, control=None)
        Fit the model on data.
        
        Parameters
        ----------
        data : pandas.DataFrame
        
        outcome : str or list of str
            Names of the outcomes.
        
        treatment : str or list of str
            Names of the treatment vectors.
        
        covariate : str of list of str
            Names of the covariate vectors.
        
        adjustment : str of list of str
            Names of the covariate vectors. Note that we may only need the covariate
            set, which usually is a subset of the adjustment set.
        
        treat : int or list, optional
            If there is only one discrete treament, then treat indicates the
            treatment group. If there are multiple treatment groups, then treat
            should be a list of str with length equal to the number of treatments. 
            For example, when there are multiple discrete treatments,
                array(['run', 'read'])
            means the treat value of the first treatment is taken as 'run' and
            that of the second treatment is taken as 'read'.
        
        control : int or list, optional
            See treat for more information

        Returns
        ----------
        instance of CausalTree
            The fitted CausalTree.

    estimate(data=None, quantity=None)
        Estimate the causal effect of the treatment on the outcome in data.
        
        Parameters
        ----------
        data : pandas.DataFrame, optional, by default None
            If None, data will be set as the training data.
        
        quantity : str, optional, by default None
            The type of the causal effect. Avaliable options are:
                
                1. 'CATE' : the estimator will evaluate the CATE;
                2. 'ATE' : the estimator will evaluate the ATE;
                3. None : the estimator will evaluate the CITE.          

        Returns
        -------
        ndarray or float, optional
            The estimated causal effect with the type of the quantity.
    
    get_depth()
        Return the depth of the causal tree. The depth of a tree is the maximum distance between the root
        and any leaf.
        
        Returns
        -------
        self.tree_.max_depth : int
            The maximum depth of the tree.
    
    get_n_leaves()
        Return the number of leaves of the causal tree.
        
        Returns
        -------
        self.tree_.n_leaves : int
            Number of leaves.
    
    apply(*, data=None, wv=None)
        Return the index of the leaf that each sample is predicted as.

        Parameters
        ----------
        wv : ndarray
            The input samples as an ndarray. If None, then the DataFrame data
            will be used as the input samples.
        
        data : DataFrame, optional
            The input samples. The data must contains columns of the covariates
            used for training the model. If None, the training data will be
            passed as input samples , by default None

        Returns
        -------
        v_leaves : array-like of shape (n_samples, )
            For each datapoint v_i in v, return the index of the leaf v_i
            ends up in. Leaves are numbered within
            ``[0; self.tree_.node_count)``, possibly with gaps in the
            numbering.

    decision_path(*, data=None, wv=None)
        Return the decision path.
        
        Parameters
        ----------
        wv : ndarray
            The input samples as an ndarray. If None, then the DataFrame data
            will be used as the input samples.
        
        data : DataFrame, optional
            The input samples. The data must contains columns of the covariates
            used for training the model. If None, the training data will be
            passed as input samples , by default None
        
        Returns
        -------
        indicator : sparse matrix of shape (n_samples, n_nodes)
            Return a node indicator CSR matrix where non zero elements
            indicates that the samples goes through the nodes.

    feature_importance()
        Returns
        -------
        feature_importances_ : ndarray of shape (n_features,)
            Normalized total reduction of criteria by feature
            (Gini importance).

    plot_causal_tree(max_depth=None, feature_names=None, label='all', filled=False, node_ids=False, proportion=False, rounded=False, precision=3, ax=None, fontsize=None)
        Plot a causal tree.
        The visualization is fit automatically to the size of the axis.
        Use the ``figsize`` or ``dpi`` arguments of ``plt.figure``  to control
        the size of the rendering.

        Parameters
        ----------        
        max_depth : int, default=None
            The maximum depth of the representation. If None, the tree is fully
            generated.
        
        label : {'all', 'root', 'none'}, default='all'
            Whether to show informative labels for impurity, etc.
            Options include 'all' to show at every node, 'root' to show only at
            the top root node, or 'none' to not show at any node.
        
        filled : bool, default=False
            When set to ``True``, paint nodes to indicate majority class for
            classification, extremity of values for regression, or purity of node
            for multi-output.

        feature_names : list, default=None
            Names of features. If None, then the names of the adjustment or covariate
            will be used.
        
        node_ids : bool, default=False
            When set to ``True``, show the ID number on each node.
        
        proportion : bool, default=False
            When set to ``True``, change the display of 'values' and/or 'samples'
            to be proportions and percentages respectively.
        
        rounded : bool, default=False
            When set to ``True``, draw node boxes with rounded corners and use
            Helvetica fonts instead of Times-Roman.
        
        precision : int, default=3
            Number of digits of precision for floating point in the values of
            impurity, threshold and value attributes of each node.
        
        ax : matplotlib axis, default=None
            Axes to plot to. If None, use current axis. Any previous content
            is cleared.
        
        fontsize : int, default=None
            Size of text font. If None, determined automatically to fit figure.
        
        Returns
        -------
        annotations : list of artists
            List containing the artists for the annotation boxes making up the
            tree.
    """
    # TODO: sample_weight

    def __init__(
        self,
        *,
        splitter='best',
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        random_state=2022,
        max_leaf_nodes=None,
        max_features=None,
        min_impurity_decrease=0.0,
        min_weight_fraction_leaf=0.0,
        ccp_alpha=0.0,
        categories='auto'
    ):
        self.splitter = splitter
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.random_state = random_state
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.min_impurity_decrease = min_impurity_decrease
        self.ccp_alpha = ccp_alpha

        super().__init__(
            random_state=random_state,
            categories=categories,
        )

    def fit(
        self,
        data,
        outcome,
        treatment,
        adjustment=None,
        covariate=None,
        treat=None,
        control=None,
    ):
        # TODO: consider possibility for generalizing continuous treatment
        """Fit the model on data.

        Parameters
        ----------
        data : pandas.DataFrame
        
        outcome : str or list of str
            Names of the outcomes.
        
        treatment : str or list of str
            Names of the treatment vectors.
        
        covariate : str of list of str
            Names of the covariate vectors.
        
        adjustment : str of list of str
            Names of the covariate vectors. Note that we may only need the covariate
            set, which usually is a subset of the adjustment set.
        
        treat : int or list, optional
            If there is only one discrete treament, then treat indicates the
            treatment group. If there are multiple treatment groups, then treat
            should be a list of str with length equal to the number of treatments. 
            For example, when there are multiple discrete treatments,
                array(['run', 'read'])
            means the treat value of the first treatment is taken as 'run' and
            that of the second treatment is taken as 'read'.
        
        control : int or list, optional
            See treat for more information

        Returns
        ----------
        instance of CausalTree
            The fitted causal tree.
        """
        assert adjustment is not None or covariate is not None, \
            'Need adjustment or covariate to perform estimation.'
        
        # check random state
        random_state = check_random_state(self.random_state)

        super().fit(
            data, outcome, treatment,
            adjustment=adjustment,
            covariate=covariate,
        )
        
        y, x, w, v = convert2array(
            data, outcome, treatment, adjustment, covariate
        )
        wv = get_wv(w, v)

        # Determin treatment settings
        if self.categories == 'auto' or self.categories is None:
            categories = 'auto'
        else:
            categories = list(self.categories)

        self.transformer = OrdinalEncoder(categories=categories)
        self.transformer.fit(x)
        x = self.transformer.transform(x)
        n_treatments = len(self.transformer.categories_)

        # get new dataset with treat and controls
        treat = get_treat_control(treat, self.transformer, True)
        control = get_treat_control(control, self.transformer, treat=False)

        self.treat = treat
        self.control = control

        # TODO: this should be much more simpler when considering single treat
        _tr = np.all(treat == x, axis=1)
        _crtl = np.all(control == x, axis=1)
        label = np.any((_tr, _crtl), axis=0)
        y = y[label]
        wv = wv[label]
        sample_weight = _tr[label].astype(int)

        check_scalar(
            self.ccp_alpha,
            name='ccp_alpha',
            target_type=numbers.Real,
            min_val=0.0,
        )        
        
        # Determine output settings
        n_samples, self.n_features_in_ = wv.shape  # dimension of the input
        self._wv = wv

        if y.ndim == 1:
            y = y.reshape(-1, 1)

        self.n_outputs_ = y.shape[1]

        # Check parameters
        if self.max_depth is not None:
            check_scalar(
                self.max_depth,
                name='max_depth',
                target_type=numbers.Integral,
                min_val=1,
            )
        max_depth = np.iinfo(np.int32).max if self.max_depth is None \
            else self.max_depth

        # check self.min_samples_leaf
        if isinstance(self.min_samples_leaf, numbers.Integral):
            check_scalar(
                self.min_samples_leaf,
                name='min_samples_leaf',
                target_type=numbers.Integral,
                min_val=1,
            )
            min_samples_leaf = self.min_samples_leaf
        else:
            check_scalar(
                self.min_samples_leaf,
                name='min_samples_leaf',
                target_type=numbers.Real,
                min_val=0.0,
                include_boundaries='neither',
            )
            min_samples_leaf = int(ceil(self.min_samples_leaf * n_samples))

        # check self.min_samples_split
        if isinstance(self.min_samples_split, numbers.Integral):
            check_scalar(
                self.min_samples_split,
                name='min_samples_split',
                target_type=numbers.Integral,
                min_val=2,
            )
            min_samples_split = self.min_samples_split
        else:
            check_scalar(
                self.min_samples_split,
                name='min_samples_split',
                target_type=numbers.Real,
                min_val=0.0,
                max_val=1.0,
                include_boundaries='right',
            )
            min_samples_split = int(ceil(self.min_samples_split * n_samples))
            min_samples_split = max(2, min_samples_split)

        min_samples_split = max(min_samples_split, 2 * min_samples_leaf)

        # check self.max_leaf_nodes
        if self.max_leaf_nodes is not None:
            check_scalar(
                self.max_leaf_nodes,
                name='max_leaf_nodes',
                target_type=numbers.Integral,
                min_val=2,
            )

        max_leaf_nodes = -1 if self.max_leaf_nodes is None else self.max_leaf_nodes

        # check min_weight_fraction_leaf
        check_scalar(
            self.min_weight_fraction_leaf,
            name='min_weight_fraction_leaf',
            target_type=numbers.Real,
            min_val=0.0,
            max_val=0.5,
        )

        # check max_features
        if isinstance(self.max_features, str):
            if self.max_features == 'sqrt':
                max_features = max(1, int(np.sqrt(self.n_features_in_)))
            elif self.max_features == 'log2':
                max_features = max(1, int(np.log2(self.n_features_in_)))
            else:
                raise ValueError(
                    'Invalid value for max_features. Allowed string values'
                    f'Allowed string values are "sqrt" or "log2", but was given {self.max_features}.'
                )
        elif self.max_features is None:
            max_features = self.n_features_in_
        elif isinstance(self.max_features, numbers.Integral):
            check_scalar(
                self.max_features,
                name='max_features',
                target_type=numbers.Integral,
                min_val=1,
                include_boundaries='left',
            )
            max_features = self.max_features
        else:
            check_scalar(
                self.max_features,
                name='max_features',
                target_type=numbers.Real,
                min_val=0.0,
                max_val=1.0,
                include_boundaries='right',
            )
            if self.max_features > 0.0:
                max_features = max(
                    1, int(self.max_features * self.n_features_in_)
                )
            else:
                max_features = 0

        self.max_features_ = max_features

        if len(y) != n_samples:
            raise ValueError(
                f'Number of labels {len(y)} does not match number of samples'
            )

        # set min_weight_leaf
        if sample_weight is None:
            min_weight_leaf = self.min_weight_fraction_leaf * n_samples
        else:
            min_weight_leaf = self.min_weight_fraction_leaf * \
                np.sum(sample_weight)

        # Build tree step 1. Set up criterion
        # criterion = deepcopy(MSE(self.n_outputs_, n_samples))
        # criterion = deepcopy(CMSE(self.n_outputs_, n_samples))
        criterion = deepcopy(HonestCMSE(self.n_outputs_, n_samples))

        logger.info(
            f'Start building the causal tree with criterion {type(criterion).__name__}'
        )

        # Build tree step 2. Define splitter
        splitter = self.splitter
        
        if not isinstance(self.splitter, Splitter):
            splitter = SPLITTERS[self.splitter](
                criterion,
                self.max_features_,
                min_samples_leaf,
                min_weight_leaf,
                random_state,
            )

        logger.info(
            f'Building the causal tree with splitter {type(splitter).__name__}'
        )

        # Build tree step 3. Define the tree
        self.tree_ = Tree(
            self.n_features_in_,
            np.array([1] * self.n_outputs_, dtype=np.intp),
            self.n_outputs_,
        )

        # Build tree step 3. Build the tree
        if max_leaf_nodes < 0:
            builder = DepthFirstTreeBuilder(
                splitter,
                min_samples_split,
                min_samples_leaf,
                min_weight_leaf,
                max_depth,
                self.min_impurity_decrease,
            )
        else:
            builder = BestFirstTreeBuilder(
                splitter,
                min_samples_split,
                min_samples_leaf,
                min_weight_leaf,
                max_depth,
                max_leaf_nodes,
                self.min_impurity_decrease,
            )

        logger.info(
            f'Building the causal tree with builder {type(builder).__name__}'
        )

        builder.build(self.tree_, wv, y, sample_weight + EPS)

        self._is_fitted = True

        return self

    def estimate(self, data=None, quantity=None):
        """Estimate the causal effect of the treatment on the outcome in the data.

        Parameters
        ----------
        data : pandas.DataFrame, optional, by default None
            If None, data will be set as the training data.
        
        quantity : str, optional, by default None
            The type of the causal effect. Avaliable options are:
                'CATE' : the estimator will evaluate the CATE;
                'ATE' : the estimator will evaluate the ATE;
                None : the estimator will evaluate the CITE.          

        Returns
        -------
        ndarray or float, optional
            The estimated causal effect with the type of the quantity.
        """
        if not self._is_fitted:
            raise Exception('The model is not fitted yet.')

        effect = self._prepare4est(data)

        logger.info(
            f'Start estimating the causal effect with the type of {quantity}.'
        )
        
        if quantity == 'ATE' or quantity == 'CATE':
            np.mean(effect, axis=0)
        else:
            return effect

    def _check_features(self, *, wv=None, data=None):
        """Validate the data for the model to estimate the causal effect.

        Parameters
        ----------
        wv : ndarray
            The test samples as an ndarray. If None, then the DataFrame data
            will be used as the test samples.
        
        data : pandas.DataFrame
            The test samples.

        Returns
        -------
        ndarray
            Valid input to the causal tree.
        """
        if wv is not None:
            wv = wv.reshape(-1, 1) if wv.ndim == 1 else wv
            assert wv.shape[1] == self.n_features_in_
            wv = wv.astype(np.float32)
            
            return wv

        if data is None:
            wv = self._wv
        else:
            assert isinstance(data, pandas.DataFrame)
            
            w, v = convert2array(data, self.adjustment, self.covariate)
            wv = get_wv(w, v)
        
        wv = wv.astype(np.float32)
        
        return wv

    def effect_nji(self, data=None):
        y_nji = self._prepare4est(data=data)
        n, y_d = y_nji.shape
        y_nji = y_nji.reshape(n, y_d, -1)
        zeros_ = np.zeros((n, y_d, 1))
        y_nji = np.concatenate((zeros_, y_nji), axis=2)
        return y_nji

    def _prepare4est(self, data=None):
        wv = self._check_features(wv=None, data=data)
        effect = self.tree_.predict(wv)

        return effect

    def _prune_tree(self):
        pass
    
    def get_depth(self):
        """Return the depth of the causal tree.
        The depth of a tree is the maximum distance between the root
        and any leaf.
        
        Returns
        -------
        self.tree_.max_depth : int
            The maximum depth of the tree.
        """
        assert self._is_fitted
        return self.tree_.max_depth

    def get_n_leaves(self):
        """Return the number of leaves of the causal tree.
        Returns
        -------
        self.tree_.n_leaves : int
            Number of leaves.
        """
        assert self._is_fitted
        return self.tree_.n_leaves


    def apply(self, *, data=None, wv=None):
        """Return the index of the leaf that each sample is predicted as.

        Parameters
        ----------
        wv : ndarray
            The input samples as an ndarray. If None, then the DataFrame data
            will be used as the input samples.
        
        data : DataFrame, optional
            The input samples. The data must contains columns of the covariates
            used for training the model. If None, the training data will be
            passed as input samples , by default None

        Returns
        -------
        v_leaves : array-like of shape (n_samples, )
            For each datapoint v_i in v, return the index of the leaf v_i
            ends up in. Leaves are numbered within
            ``[0; self.tree_.node_count)``, possibly with gaps in the
            numbering.
        """
        assert self._is_fitted, 'The model is not fitted yet.'
        wv = self._check_features(wv=wv, data=data)
        
        return self.tree_.apply(wv)     

    def decision_path(self, *, data=None, wv=None):
        """Return the decision path in the tree.

        Parameters
        ----------
        wv : ndarray
            The input samples as an ndarray. If None, then the DataFrame data
            will be used as the input samples.
        
        data : DataFrame, optional
            The input samples. The data must contains columns of the covariates
            used for training the model. If None, the training data will be
            passed as input samples , by default None
        
        Returns
        -------
        indicator : sparse matrix of shape (n_samples, n_nodes)
            Return a node indicator CSR matrix where non zero elements
            indicates that the samples goes through the nodes.
        """
        assert self._is_fitted, 'The model is not fitted yet.'

        v = self._check_features(wv=wv, data=data)

        return self.tree_.decision_path(v)
    
    @property
    def feature_importance(self):
        """Return the feature importances.
        The importance of a feature is computed as the (normalized) total
        reduction of the criterion brought by that feature.
        It is also known as the Gini importance.
        Warning: impurity-based feature importances can be misleading for
        high cardinality features (many unique values). See
        :func:`sklearn.inspection.permutation_importance` as an alternative.

        Returns
        -------
        feature_importances_ : ndarray of shape (n_features,)
            Normalized total reduction of criteria by feature
            (Gini importance).
        """
        assert self._is_fitted, 'The model is not fitted yet.'
        
        return self.tree_.compute_feature_importances()
    
    @property
    def n_features_(self):
        return self.n_features_in_
    
    def plot_causal_tree(
        self, *,
        max_depth=None,
        feature_names=None,
        label='all',
        filled=False,
        node_ids=False,
        proportion=False,
        rounded=False,
        precision=3,
        ax=None,
        fontsize=None
    ):
        """Plot a causal tree.
        The visualization is fit automatically to the size of the axis.
        Use the ``figsize`` or ``dpi`` arguments of ``plt.figure``  to control
        the size of the rendering.

        Parameters
        ----------        
        max_depth : int, default=None
            The maximum depth of the representation. If None, the tree is fully
            generated.
        
        label : {'all', 'root', 'none'}, default='all'
            Whether to show informative labels for impurity, etc.
            Options include 'all' to show at every node, 'root' to show only at
            the top root node, or 'none' to not show at any node.
        
        filled : bool, default=False
            When set to ``True``, paint nodes to indicate majority class for
            classification, extremity of values for regression, or purity of node
            for multi-output.

        feature_names : list, default=None
            Names of features. If None, then the names of the adjustment or covariate
            will be used.
        
        node_ids : bool, default=False
            When set to ``True``, show the ID number on each node.
        
        proportion : bool, default=False
            When set to ``True``, change the display of 'values' and/or 'samples'
            to be proportions and percentages respectively.
        
        rounded : bool, default=False
            When set to ``True``, draw node boxes with rounded corners and use
            Helvetica fonts instead of Times-Roman.
        
        precision : int, default=3
            Number of digits of precision for floating point in the values of
            impurity, threshold and value attributes of each node.
        
        ax : matplotlib axis, default=None
            Axes to plot to. If None, use current axis. Any previous content
            is cleared.
        
        fontsize : int, default=None
            Size of text font. If None, determined automatically to fit figure.
        
        Returns
        -------
        annotations : list of artists
            List containing the artists for the annotation boxes making up the
            tree.
        """
        assert self._is_fitted

        impurity = False

        if feature_names is None:
            feature_names = []
            
            if self.adjustment is not None:
                if isinstance(self.adjustment, str):
                    adj = [self.adjustment]
                else:
                    adj = list(self.adjustment)
                feature_names.extend(adj)
            
            if self.covariate is not None:
                if isinstance(self.covariate, str):
                    cov = [self.covariate]
                else:
                    cov = list(self.covariate)
                feature_names.extend(cov)

        return plot_tree(
            self,
            max_depth=max_depth,
            feature_names=feature_names,
            class_names=None,
            label=label,
            filled=filled,
            impurity=impurity,
            node_ids=node_ids,
            proportion=proportion,
            rounded=rounded,
            precision=precision,
            ax=ax,
            fontsize=fontsize,
        )

# class _CausalTreeOld:
#     """
#     A class for estimating causal effect with decision tree.

#     Attributes
#     ----------
#     feature_importances_ : ndarray of shape (n_features,)
#         The feature importances.
#     tree_ : Tree instance
#         The underlying Tree object.
#     max_depth : int, by default None
#     min_samples_split : int or float, by default 2
#     min_samples_leaf : int or float, by default 1
#     random_state : int
#     max_leaf_nodes : int, by default None
#     min_impurity_decrease : float, by default 0.0
#     ccp_alpha : non-negative float, by default 0.0

#     Methods
#     ----------
#     fit()
#     predict()
#     estimate()
#     apply()
#     """

#     def __init__(
#         self,
#         max_depth=None,
#         min_samples_split=2,
#         min_samples_leaf=1,
#         random_state=None,
#         max_leaf_nodes=None,
#         min_impurity_decrease=0.0,
#         ccp_alpha=0.0,
#     ):
#         """
#         Many parameters are similar to those of BaseDecisionTree of sklearn.

#         Parameters
#         ----------
#         max_depth : int, by default None
#             The maximum depth of the tree.
#         min_samples_split : int or float, by default 2
#             The minimum number of samples required to split an internal node.
#         min_samples_leaf : int or float, by default 1
#             The minimum number of samples required to be at a leaf node.
#         random_state : int
#         max_leaf_nodes : int, by default None
#             Grow a tree with ``max_leaf_nodes`` in best-first fashion.
#             Best nodes are defined as relative reduction in impurity.
#             If None then unlimited number of leaf nodes.
#         min_impurity_decrease : float, by default 0.0
#             A node will be split if this split induces a decrease of the
#             impurity greater than or equal to this value.
#         ccp_alpha : non-negative float, by default 0.0
#             Value for pruning the tree. #TODO: not implemented yet.

#         See Also
#         --------
#         BaseDecisionTree : The default implementation of decision tree.
#         """
#         self.max_depth = max_depth
#         self.min_samples_split = min_samples_split
#         self.min_samples_leaf = min_samples_leaf
#         self.random_state = random_state
#         self.max_leaf_nodes = max_leaf_nodes
#         self.min_impurity_decrease = min_impurity_decrease
#         self.ccp_alpha = ccp_alpha
#         self.tree_ = None

#     def fit(
#         self,
#         data,
#         outcome,
#         treatment,
#         adjustment=None,
#         covariate=None,
#     ):
#         """Fit the model to data.

#         Parameters
#         ----------
#         data : pandas.DataFrame
#             Currently only support this.
#         outcome : str or list of str
#             Name of the outcome.
#         treatment : str or list of str
#             Name of the treatment vector.
#         treatment_value : int, optional, defaults to 1
#             If treatment == treatment_value for a data point, then it is in
#             the treatment group, otherwise it is in the control group.

#         Returns:
#             self: The fitted causal tree.
#         """
#         assert adjustment is not None or covariate is not None, \
#             'Need adjustment or covariate to perform estimation.'

#         # Determine output settings
#         self.outcome = outcome
#         self.treatment = treatment
#         self.adjustment = adjustment
#         self.covariate = covariate

#         n = len(data)
#         y, x, w, v = convert2array(
#             data, outcome, treatment, adjustment, covariate
#         )

#         if y.ndim == 1:
#             y = y.reshape(-1, 1)
#         n_samples, self.n_features_in_ = X.shape  # dimension of the input
#         n_outputs = 1

#         # Check parameters and etermine output settings
#         sample_weight = (t == treatment_value).astype(int)
#         min_weight_leaf = 2  # maybe this needs more modifications

#         # Build tree step 1. Set up criterion
#         criterion = CMSE(1, n_samples)

#         # Build tree step 2. Define splitter
#         splitter = BestSplitter(
#             criterion,
#             self.n_features_in_,
#             self.min_samples_leaf,
#             min_weight_leaf,
#             self.random_state,
#         )

#         # Build tree step 3. Define the tree
#         self.tree_ = Tree(
#             self.n_features_in_,
#             np.array([1] * n_outputs, dtype=np.intp),
#             n_outputs,
#         )

#         # Build tree step 3. Build the tree
#         # TODO: try to be more compatible with the sklearn BasBaseDecisionTree
#         builder = DepthFirstTreeBuilder(
#             splitter,
#             self.min_samples_split,
#             self.min_samples_leafmin_samples_leaf,
#             min_weight_leaf,
#             self.max_depth,
#             self.min_impurity_decrease,
#         )
#         builder.build(self.tree_, X, y, sample_weight)

#         return self

#     def predict(self, data, outcome=None, treatment=None):
#         if treatment is not None:
#             data = data.drop(treatment, axis=1)
#         if outcome is not None:
#             data = data.drop(outcome, axis=1)

#         check_is_fitted(self)
#         proba = self.tree_.predict(data)
#         return proba[:, 0]

#     def estimate(self, data, outcome, treatment, treatment_value=1):
#         """Based on my current understanding, the causal tree only solves
#         estimation of heterogeneous causal effects. Thus we may not need
#         estimations for other effects.
#         """
#         self.fit(data, outcome, treatment, treatment_value)
#         result = self.predict(data, outcome, treatment)
#         return result

#     def _prune_tree(self):
#         pass

#     def apply(self, X):
#         """Return the index of the leaf that each sample is predicted as.
#         """
#         check_is_fitted(self)
#         return self.tree_.apply(X)

#     @property
#     def feature_importance(self):
#         pass



"""Utilities for input validation"""

# Authors: Olivier Grisel
#          Gael Varoquaux
#          Andreas Mueller
#          Lars Buitinck
#          Alexandre Gramfort
#          Nicolas Tresegnie
#          Sylvain Marie
# License: BSD 3 clause

from functools import wraps
import warnings
import numbers
import operator

import numpy as np


def check_random_state(seed):
    """Turn seed into a np.random.RandomState instance.

    Parameters
    ----------
    seed : None, int or instance of RandomState
        If seed is None, return the RandomState singleton used by np.random.
        If seed is an int, return a new RandomState instance seeded with seed.
        If seed is already a RandomState instance, return it.
        Otherwise raise ValueError.

    Returns
    -------
    :class:`numpy:numpy.random.RandomState`
        The random state object based on `seed` parameter.
    """
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, numbers.Integral):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError(
        "%r cannot be used to seed a numpy.random.RandomState instance" % seed
    )




import networkx as nx


class CausalStructuralModel:
    def __init__(self, causation):
        self.edges = self.build_edges(causation)
        pass

    def build_edges(self, causation):
        pass





class Do:
    """Implementation of the do-operator.

    Attributes
    ----------
    causation : CausalGraph or CausalStructureModel
    treatment : str

    Methods
    ----------
    """

    def __init__(self, causation):
        """
        Parameters
        ----------
        causation : CausalGraph or CausalStructureModel
            the causation structures in which the do-operator will be applied
        """
        self.causation = causation

    def do(self, treatment, value):
        """Do the treatment=value in the graph or structural models.

        Parameters
        ----------
        treatment : str
        value : float, optional
            the value of the treatment
        """
        self.causation.remove_incoming_edges(treatment)
        # TODO

    # def remove_incoming_edges(self, node):
    #     """remove the incoming edges to node

    #     Parameters
    #     ----------
    #     node : str
    #     """
    #     pass


class DoCalculus:
    # TODO: Rules of do-calculus are implemented as follows
    def rule_one(self):
        pass

    def rule_two(self):
        pass

    def rule_three(self):
        pass



# -*- coding:utf-8 -*-
"""
Clone from Hypernets: https://github.com/DataCanvasIO/Hypernets/blob/master/hypernets/tabular/dataframe_mapper.py
"""
import hashlib

import numpy as np
import pandas as pd
from scipy import sparse as _sparse
from sklearn.base import BaseEstimator
from sklearn.pipeline import _name_estimators, Pipeline
from sklearn.utils import tosequence

from ylearn.utils import logging, context

logger = logging.get_logger(__name__)


def _call_fit(fit_method, X, y=None, **kwargs):
    """
    helper function, calls the fit or fit_transform method with the correct
    number of parameters

    fit_method: fit or fit_transform method of the transformer
    X: the data to fit
    y: the target vector relative to X, optional
    kwargs: any keyword arguments to the fit method

    return: the result of the fit or fit_transform method

    WARNING: if this function raises a TypeError exception, test the fit
    or fit_transform method passed to it in isolation as _call_fit will not
    distinguish TypeError due to incorrect number of arguments from
    other TypeError
    """
    try:
        return fit_method(X, y, **kwargs)
    except TypeError:
        # fit takes only one argument
        return fit_method(X, **kwargs)


def _hash(data):
    m = hashlib.md5()
    m.update(data)
    return m.hexdigest()


class TransformerPipeline(Pipeline):
    """
    Pipeline that expects all steps to be transformers taking a single X
    argument, an optional y argument, and having fit and transform methods.

    Code is copied from sklearn's Pipeline
    """

    def __init__(self, steps):
        names, estimators = zip(*steps)
        if len(dict(steps)) != len(steps):
            raise ValueError(
                "Provided step names are not unique: %s" % (names,))

        # shallow copy of steps
        self.steps = tosequence(steps)
        estimator = estimators[-1]

        for e in estimators:
            if (not (hasattr(e, "fit") or hasattr(e, "fit_transform")) or not
            hasattr(e, "transform")):
                raise TypeError("All steps of the chain should "
                                "be transforms and implement fit and transform"
                                " '%s' (type %s) doesn't)" % (e, type(e)))

        if not hasattr(estimator, "fit"):
            raise TypeError("Last step of chain should implement fit "
                            "'%s' (type %s) doesn't)"
                            % (estimator, type(estimator)))

    def _pre_transform(self, X, y=None, **fit_params):
        fit_params_steps = dict((step, {}) for step, _ in self.steps)
        for pname, pval in fit_params.items():
            step, param = pname.split('__', 1)
            fit_params_steps[step][param] = pval
        Xt = X
        for name, transform in self.steps[:-1]:
            if hasattr(transform, "fit_transform"):
                Xt = _call_fit(transform.fit_transform,
                               Xt, y, **fit_params_steps[name])
            else:
                Xt = _call_fit(transform.fit,
                               Xt, y, **fit_params_steps[name]).transform(Xt)
        return Xt, fit_params_steps[self.steps[-1][0]]

    def fit(self, X, y=None, **fit_params):
        Xt, fit_params = self._pre_transform(X, y, **fit_params)
        _call_fit(self.steps[-1][-1].fit, Xt, y, **fit_params)
        return self

    def fit_transform(self, X, y=None, **fit_params):
        Xt, fit_params = self._pre_transform(X, y, **fit_params)
        if hasattr(self.steps[-1][-1], 'fit_transform'):
            return _call_fit(self.steps[-1][-1].fit_transform,
                             Xt, y, **fit_params)
        else:
            return _call_fit(self.steps[-1][-1].fit,
                             Xt, y, **fit_params).transform(Xt)


def make_transformer_pipeline(*steps):
    """Construct a TransformerPipeline from the given estimators.
    """
    return TransformerPipeline(_name_estimators(steps))


def _get_feature_names(estimator, columns=None):
    """
    Attempt to extract feature names based on a given estimator
    """
    if hasattr(estimator, 'get_feature_names'):
        return estimator.get_feature_names(columns)
    if hasattr(estimator, 'classes_'):
        return estimator.classes_
    return None


class DataFrameMapper(BaseEstimator):
    """
    Map Pandas data frame column subsets to their own sklearn transformation.

    Parameters:
    ----------
    features :  a list of tuples with features definitions.
                The first element is the pandas column selector. This can
                be a string (for one column) or a list of strings.
                The second element is an object that supports
                sklearn's transform interface, or a list of such objects.
                The third element is optional and, if present, must be
                a dictionary with the options to apply to the
                transformation. Example: {'alias': 'day_of_week'}

    default :   default transformer to apply to the columns not
                explicitly selected in the mapper. If False (default),
                discard them. If None, pass them through untouched. Any
                other transformer will be applied to all the unselected
                columns as a whole, taken as a 2d-array.

    df_out :    return a pandas data frame, with each column named using
                the pandas column that created it (if there's only one
                input and output) or the input columns joined with '_'
                if there's multiple inputs, and the name concatenated with
                '_1', '_2' etc if there's multiple outputs.

    input_df :  If ``True`` pass the selected columns to the transformers
                as a pandas DataFrame or Series. Otherwise pass them as a
                numpy array. Defaults to ``False``.

    Attributes
    ----------
    fitted_features_ : list of tuple(column_name list, fitted transformer, options).
    """

    def __init__(self, features, default=False, df_out=False, input_df=False, df_out_dtype_transforms=None):
        self.features = features
        self.default = default
        self.df_out = df_out
        self.input_df = input_df
        self.df_out_dtype_transforms = df_out_dtype_transforms

        # fitted
        self.feature_names_in_ = None
        self.n_features_in_ = None
        self.fitted_features_ = None

    @staticmethod
    def _build_transformer(transformers):
        if isinstance(transformers, list):
            transformers = make_transformer_pipeline(*transformers)
        return transformers

    def _build_feature(self, columns, transformers, options={}):
        return columns, self._build_transformer(transformers), options

    def _build(self, features, default):
        if isinstance(features, list):
            built_features = [self._build_feature(*f) for f in features]
        else:
            built_features = features
        built_default = self._build_transformer(default)

        return built_features, built_default

    def fit(self, X, y=None):
        built_features, built_default = self._build(self.features, self.default)

        columns_in = X.columns.to_list()
        fitted_features = []
        selected_columns = []

        for columns_def, transformers, options in built_features:
            logger.debug(f'columns:({columns_def}), transformers:({transformers}), options:({options})')
            if callable(columns_def):
                columns = columns_def(X)
            elif isinstance(columns_def, str):
                columns = [columns_def]
            else:
                columns = columns_def

            if isinstance(columns, (list, tuple)):
                columns = [c for c in columns if c not in selected_columns]

            fitted_features.append((columns, transformers, options))
            if columns is None or len(columns) <= 0:
                continue

            selected_columns += columns
            if transformers is not None:
                input_df = options.get('input_df', self.input_df)
                with context(columns):
                    Xt = self._get_col_subset(X, columns, input_df)
                    if logger.is_debug_enabled():
                        msg = f'fit {type(transformers).__name__}, X shape:{X.shape}'
                        if hasattr(transformers, 'steps'):
                            msg += f", steps:{[s[0] for s in transformers.steps]}"
                        logger.debug(msg)
                    _call_fit(transformers.fit, Xt, y)
                    # print(f'{transformers}:{Xt.dtypes}')

        # handle features not explicitly selected
        if built_default is not False and len(X.columns) > len(selected_columns):
            unselected_columns = [c for c in X.columns.to_list() if c not in selected_columns]
            if built_default is not None:
                with context(unselected_columns):
                    Xt = self._get_col_subset(X, unselected_columns, self.input_df)
                    _call_fit(built_default.fit, Xt, y)
            fitted_features.append((unselected_columns, built_default, {}))

        self.feature_names_in_ = columns_in
        self.n_features_in_ = len(columns_in)
        self.fitted_features_ = fitted_features

        return self

    def transform(self, X):
        selected_columns = []
        transformed_columns = []
        extracted = []

        for columns, transformers, options in self.fitted_features_:
            if columns is None or len(columns) < 1:
                continue
            selected_columns += columns

            input_df = options.get('input_df', self.input_df)
            alias = options.get('alias')

            Xt = self._get_col_subset(X, columns, input_df)
            if transformers is not None:
                with context(columns):
                    if logger.is_debug_enabled():
                        msg = f'transform {type(transformers).__name__}, X shape:{X.shape}'
                        if hasattr(transformers, 'steps'):
                            msg += f", steps:{[s[0] for s in transformers.steps]}"
                        logger.debug(msg)
                    # print(f'before ---- {transformers}:{Xt.dtypes}')
                    Xt = transformers.transform(Xt)
                    # print(f'after ---- {transformers}:{pd.DataFrame(Xt).dtypes}')

            extracted.append(self._fix_feature(Xt))
            transformed_columns += self._get_names(columns, transformers, Xt, alias)

        return self._to_transform_result(X, extracted, transformed_columns)

    def fit_transform(self, X, y=None, *fit_args):
        columns_in = X.columns.to_list()
        fitted_features = []
        selected_columns = []
        transformed_columns = []
        extracted = []

        built_features, built_default = self._build(self.features, self.default)
        for columns_def, transformers, options in built_features:
            if callable(columns_def):
                columns = columns_def(X)
            elif isinstance(columns_def, str):
                columns = [columns_def]
            else:
                columns = columns_def
            if isinstance(columns, (list, tuple)) and len(set(selected_columns).intersection(set(columns))) > 0:
                columns = [c for c in columns if c not in selected_columns]

            if columns is None or len(columns) < 1:
                continue

            fitted_features.append((columns, transformers, options))
            selected_columns += columns
            if logger.is_debug_enabled():
                logger.debug(f'fit_transform {len(columns)} columns with:\n{transformers}')

            input_df = options.get('input_df', self.input_df)
            alias = options.get('alias')

            Xt = self._get_col_subset(X, columns, input_df)
            if transformers is not None:
                with context(columns):
                    if logger.is_debug_enabled():
                        msg = f'fit_transform {type(transformers).__name__}, X shape:{X.shape}'
                        if hasattr(transformers, 'steps'):
                            msg += f", steps:{[s[0] for s in transformers.steps]}"
                        logger.debug(msg)
                    if hasattr(transformers, 'fit_transform'):
                        Xt = _call_fit(transformers.fit_transform, Xt, y)
                    else:
                        _call_fit(transformers.fit, Xt, y)
                        Xt = transformers.transform(Xt)

            extracted.append(self._fix_feature(Xt))
            if logger.is_debug_enabled():
                logger.debug(f'columns:{len(columns)}')
            transformed_columns += self._get_names(columns, transformers, Xt, alias)
            if logger.is_debug_enabled():
                logger.debug(f'transformed_names_:{len(transformed_columns)}')

        # handle features not explicitly selected
        if built_default is not False and len(X.columns) > len(selected_columns):
            unselected_columns = [c for c in X.columns.to_list() if c not in selected_columns]
            Xt = self._get_col_subset(X, unselected_columns, self.input_df)
            if built_default is not None:
                with context(unselected_columns):
                    if hasattr(built_default, 'fit_transform'):
                        Xt = _call_fit(built_default.fit_transform, Xt, y)
                    else:
                        _call_fit(built_default.fit, Xt, y)
                        Xt = built_default.transform(Xt)
                transformed_columns += self._get_names(unselected_columns, built_default, Xt)
            else:
                # if not applying a default transformer, keep column names unmodified
                transformed_columns += unselected_columns
            extracted.append(self._fix_feature(Xt))

            fitted_features.append((unselected_columns, built_default, {}))

        self.feature_names_in_ = columns_in
        self.n_features_in_ = len(columns_in)
        self.fitted_features_ = fitted_features

        return self._to_transform_result(X, extracted, transformed_columns)

    @staticmethod
    def _get_col_subset(X, cols, input_df=False):
        t = X[cols]
        if input_df:
            return t
        else:
            return t.values

    def _get_names(self, columns, transformer, x, alias=None):
        """
        Return verbose names for the transformed columns.

        columns       name (or list of names) of the original column(s)
        transformer   transformer - can be a TransformerPipeline
        x             transformed columns (numpy.ndarray)
        alias         base name to use for the selected columns
        """
        # logger.debug(
        #     f'get_names: {isinstance(columns, list)}, len(columns):{len(columns)} columns:{columns}, alias:{alias}')
        if alias is not None:
            name = alias
        elif isinstance(columns, list):
            name = '_'.join(map(str, columns))
            if len(name) > 64:
                name = name[:32] + _hash(name.encode('utf-8'))
        else:
            name = columns
        num_cols = x.shape[1] if len(x.shape) > 1 else 1
        if num_cols > 1:
            # If there are as many columns as classes in the transformer,
            # infer column names from classes names.

            # If we are dealing with multiple transformers for these columns
            # attempt to extract the names from each of them, starting from the
            # last one
            # logger.debug(f'transformer:{transformer}')
            if isinstance(transformer, (TransformerPipeline, Pipeline)):
                inverse_steps = transformer.steps[::-1]
                # estimators = (estimator for _, estimator in inverse_steps)
                # names_steps = (_get_feature_names(e, columns) for e in estimators)
                # names = next((n for n in names_steps if n is not None), None)
                names = None
                for _, estimator in inverse_steps:
                    names = _get_feature_names(estimator, columns)
                    if names is not None and len(names) == num_cols:
                        break
            else:  # Otherwise use the only estimator present
                names = _get_feature_names(transformer, columns)

            if names is None and len(columns) == num_cols:
                names = list(columns)

            if names is not None and len(names) == num_cols:
                names = list(names)  # ['%s_%s' % (name, o) for o in names]
            else:  # otherwise, return name concatenated with '_1', '_2', etc.
                names = [name + '_' + str(o) for o in range(num_cols)]

            if logger.is_debug_enabled():
                # logger.debug(f'names:{names}')
                logger.debug(f'transformed names:{len(names)}')
            return names
        else:
            return [name]

    @staticmethod
    def _fix_feature(fea):
        if _sparse.issparse(fea):
            fea = fea.toarray()

        if len(fea.shape) == 1:
            """
            Convert 1-dimensional arrays to 2-dimensional column vectors.
            """
            fea = np.array([fea]).T

        return fea

    def _to_transform_result(self, X, extracted, transformed_columns):
        if extracted is None or len(extracted) == 0:
            raise ValueError("No data output, ??? ")

        if self.df_out:
            df = self._to_df(X, extracted, transformed_columns)
            df = self._dtype_transform(df)
            return df
        else:
            return self._hstack_array(extracted)

    @staticmethod
    def _hstack_array(extracted):
        stacked = np.hstack(extracted)
        return stacked

    def _to_df(self, X, extracted, columns):
        dfs = [pd.DataFrame(arr, index=None).reset_index(drop=True) for arr in extracted]
        df = pd.concat(dfs, axis=1, ignore_index=True) if len(dfs) > 1 else dfs[0]
        df.columns = columns
        if len(X) == len(df):
            df.index = X.index  # reuse the original index

        return df

    def _dtype_transform(self, df_out):
        if self.df_out_dtype_transforms is not None:
            for columns, dtype in self.df_out_dtype_transforms:
                if callable(columns):
                    columns = columns(df_out)
                if isinstance(columns, list) and len(columns) <= 0:
                    continue
                df_out[columns] = df_out[columns].astype(dtype)
        return df_out



import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data import DataLoader

from ylearn.utils import to_repr
from .utils import BatchData
# from ..utils._common import check_cols

#TODO: add support for assigning different treatment values for different examples for all models.

class BaseEstModel:
    """
    Base class for various estimation learner.

    Attributes
    ----------
    ml_model_dic : dict
        A dictionary of default machine learning sklearn models currently
        including
            'LR': LinearRegression
            'LogistR': LogisticRegression.

    Methods
    ----------
    _prepare(data, outcome, treatment, adjustment, individual=None)
        Prepare (fit the model) for estimating various quantities including
        ATE, CATE, ITE, and CITE.
    estimate(data, outcome, treatment, adjustment, quantity='ATE',
                 condition_set=None, condition=None, individual=None)
        Integrate estimations for various quantities into a single method.
    estimate_ate(self, data, outcome, treatment, adjustment)
    estimate_cate(self, data, outcome, treatment, adjustment,
                      condition_set, condition)
    estimate_ite(self, data, outcome, treatment, adjustment, individual)
    estimate_cite(self, data, outcome, treatment, adjustment,
                      condition_set, condition, individual)
    """

    def __init__(
        self,
        random_state=2022,
        is_discrete_treatment=False,
        is_discrete_outcome=False,
        _is_fitted=False,
        # is_discrete_instrument=False,
        categories='auto'
    ):
        self.random_state = random_state if random_state is not None else 2022
        self.is_discrete_treatment = is_discrete_treatment
        self.is_discrete_outcome = is_discrete_outcome
        self.categories = categories

        self._is_fitted = _is_fitted

        # fitted
        self.treatment = None
        self.outcome = None
        self.treats_ = None

    def fit(
        self,
        data,
        outcome,
        treatment,
        # adjustment=None,
        # covariate=None,
        **kwargs,
    ):
        assert data is not None and isinstance(data, pd.DataFrame)

        # check_cols(data, treatment, outcome)
        if isinstance(treatment, str):
            treatment = [treatment]

        self.treatment = treatment
        self.outcome = outcome

        for k, v in kwargs.items():
            setattr(self, k, v)
            # check_cols(data, v)

        if data is not None and self.is_discrete_treatment:
            treats = {}
            for t in treatment:
                treats[t] = tuple(np.sort(data[t].unique()).tolist())
        else:
            treats = None
        self.treats_ = treats

        return self

    def effect_nji(self, *args, **kwargs):
        raise NotImplementedError()

    #
    # def _prepare_(
    #     self,
    #     data,
    # ):
    #     pass
    #
    # def _prepare(
    #     self,
    #     data,
    #     outcome,
    #     treatment,
    #     adjustment,
    #     individual=None,
    #     **kwargs
    # ):
    #     # TODO: We should add a method like check_is_fitted.
    #     # This does not need the parameters like treatment. By calling this
    #     # function first, we then estimate quantites like ATE easily.
    #     r"""Prepare (fit the model) for estimating the quantities
    #         ATE: E[y|do(x_1)] - E[y|do(x_0)] = E_w[E[y|x=x_1,w] - E[y|x=x_0, w]
    #                                        := E_{adjustment}[
    #                                            Delta E[outcome|treatment,
    #                                                             adjustment]]
    #         CATE: E[y|do(x_1), z] - E[y|do(x_0), z] = E_w[E[y|x=x_1, w, z] -
    #                                                     E[y|x=x_0, w, z]]
    #         ITE: y_i(do(x_1)) - y_i(do(x_0))
    #         CITE: y_i(do(x_1))|z_i - y_i(do(x_0))|z_i
    #
    #     Parameters
    #     ----------
    #     data : DataFrame
    #     outcome : str
    #         Name of the outcome.
    #     treatment : str
    #         Name of the treatment.
    #     adjustment : set or list
    #         The adjutment set for the causal effect,
    #         i.e., P(outcome|do(treatment)) =
    #             \sum_{adjustment} P(outcome|treatment, adjustment)P(adjustment)
    #     individual : DataFrame, default to None
    #         The individual data for computing its causal effect.
    #
    #     Returns
    #     ----------
    #     np.array
    #     """
    #     pass

    def estimate(self, data=None, **kwargs):
        if not hasattr(self, '_is_fitted'):
            raise Exception('The estimator has not been fitted yet.')
    #
    # # def estimate(
    # #     self,
    # #     data,
    # #     outcome,
    # #     treatment,
    # #     adjustment,
    # #     quantity='ATE',
    # #     condition_set=None,
    # #     condition=None,
    # #     individual=None,
    # #     **kwargs
    # # ):
    # #     # This is the basic API for estimating a causal effect.
    #
    # #     # TODO: Note that we require that if parameters like treatment change
    # #     # then we should recall self.fit()
    # #     # method.
    # #     """General estimation method for quantities like ATE of
    # #     outcome|do(treatment).
    #
    # #     Parameter
    # #     ----------
    # #     data : DataFrame
    # #     outcome : str
    # #         Name of the outcome.
    # #     treatment : str
    # #         Name of the treatment.
    # #     adjustment : set or list
    # #         The valid adjustment set.
    # #     quantity: str
    # #         The type of desired quantity, including ATE, CATE, ITE and
    # #         CITE. Defaults to 'ATE'.
    # #     condition_set : set or list. Defaults to None
    # #     condition : list
    # #         A list whose length is the size of dataset and elements are
    # #         boolean such that we only perform the computation of
    # #         quantities if the corresponding element is True]. Defaults
    # #         to None.
    # #     individual : DataFrame. Defaults to None
    # #     kwargs : dict
    #
    # #     Raises
    # #     ----------
    # #     Exception
    # #         Raise exception if the quantity is not in ATE, CATE, ITE or CITE.
    #
    # #     Returns
    # #     ----------
    # #     float
    # #         The desired causal effect.
    # #     """
    # #     if quantity == 'ATE':
    # #         return self.estimate_ate(
    # #             data, outcome, treatment, adjustment, **kwargs
    # #         )
    # #     elif quantity == 'CATE':
    # #         return self.estimate_cate(
    # #             data, outcome, treatment, adjustment, condition_set,
    # #             condition, **kwargs
    # #         )
    # #     elif quantity == 'ITE':
    # #         return self.estimate_ite(
    # #             data, outcome, treatment, adjustment, individual, **kwargs
    # #         )
    # #     elif quantity == 'CITE':
    # #         return self.estimate_cite(
    # #             data, outcome, treatment, adjustment, condition_set,
    # #             condition, **kwargs
    # #         )
    # #     else:
    # #         raise Exception(
    # #             'Do not support estimation of quantities other'
    # #             'than ATE, CATE, ITE, or CITE'
    # #         )
    #
    # def estimate_ate(self, data, outcome, treatment, adjustment, **kwargs):
    #     """Estimate E[outcome|do(treatment=x1) - outcome|do(treatment=x0)]
    #
    #     Parameters
    #     ----------
    #     data : DataFrame
    #     outcome : str
    #         Name of the outcome.
    #     treatment : str
    #         Name of the treatment.
    #     adjustment : set or list of str
    #         The valid adjustment set.
    #
    #     Returns
    #     ----------
    #     float
    #     """
    #     return self._prepare(
    #         data, outcome, treatment, adjustment=adjustment, **kwargs
    #     ).mean()
    #
    # def estimate_cate(self, data, outcome, treatment, adjustment,
    #                   condition_set, condition, **kwargs):
    #     """Estimate E[outcome|do(treatment=x1), condition_set=condition
    #                 - outcome|do(treatment=x0), condition_set=condition]
    #
    #     Parameters
    #     ----------
    #     data : DataFrame
    #     outcome : str
    #         Name of the outcome.
    #     treatment : str
    #         Name of the treatment.
    #     adjustment : set or list of str
    #         The valid adjustment set.
    #     condition_set : set of str
    #     condition : boolean
    #         The computation will be performed only using data where conition is
    #         True.
    #
    #     Returns
    #     ----------
    #     float
    #     """
    #     # TODO: considering let the final model also be the function of
    #     # conditional variables.
    #     # modify the data for estmating cate.
    #     assert condition_set is not None, \
    #         'Need an explicit condition set to perform the analysis.'
    #
    #     new_data = data.loc[condition].drop(list(condition_set), axis=1)
    #     cate = self.estimate_ate(
    #         new_data, outcome, treatment, adjustment, **kwargs
    #     )
    #     return cate
    #
    # def estimate_ite(self, data, outcome, treatment, adjustment,
    #                  individual, **kwargs):
    #     assert individual is not None, \
    #         'Need an explicit individual to perform computation of individual'
    #     'causal effect.'
    #
    #     return self._prepare(
    #         data, outcome, treatment, adjustment,
    #         individual=individual, **kwargs
    #     )
    #
    # def estimate_cite(self, data, outcome, treatment, adjustment,
    #                   condition_set, condition, individual, **kwargs):
    #     assert individual is not None, \
    #         'Need an explicit individual to perform computation of individual'
    #     'causal effect.'
    #     assert condition_set is not None, \
    #         'Need an explicit condition set to perform the analysis.'
    #
    #     new_data = data.loc[condition].drop(list(condition_set), axis=1)
    #     cite = self._prepare(
    #         new_data, outcome, treatment, adjustment, individual, **kwargs
    #     )
    #     return cite
    #
    # def counterfactual_prediction(self):
    #     pass

    def __repr__(self):
        return to_repr(self)


class MLModel:
    """
    A parent class for possible new machine learning models which are not
    supported by sklearn.
    """

    def __init__(self, model):
        """
        Parameters
        ----------
        model : nn, optional
            This can be any machine learning models.
        """
        self.model = model

    def fit(self, X, y, nn_torch=True, **kwargs):
        if nn_torch:
            self._fit_nn_torch(X, y, **kwargs)
        else:
            # define this for other types of fit functions
            pass

    def _fit_nn_torch(
        self,
        X, y,
        device='cuda',
        lr=0.01,
        epoch=1000,
        optimizer='SGD',
        batch_size=64,
        **optim_config
    ):
        """Train the nn model with data (X, y).

        Parameters
        ----------
        X : tensor
            Has shape (b, in_d) where b is the batch size or the number of data
            points and in_d is the dimension of each data point.
        y : tensor
            Has shape (b, out_d) where out_d is the dimension of each y.
        device : str, optional. Defaults to 'cuda'.
        lr : float, optional. Defaults to 0.01
            Learning rate.
        epoch : int, optional. Defaults to 1000
            The number of epochs used for training.
        optimizer : str, optional. Defaults to 'SGD'
            Currently including SGD and Adam The type of optimizer used for
            training.
        batch_size: int, optional. Defaults to 128
        optim_config : other parameters for various optimizers.
        """
        self.model = self.model.to(device)
        op = {
            'SGD': optim.SGD(self.model.parameters(), lr=lr),
            'Adam': optim.Adam(self.model.parameters(), lr=lr, **optim_config)
        }
        opt = op[optimizer]
        loss_fn = optim_config['loss']
        data = BatchData(X=X, y=y)
        train_loader = DataLoader(data, batch_size=batch_size)

        for e in range(epoch):
            for i, (X, y) in enumerate(train_loader):
                self.model.train()
                X, y = X.to(device), y.to(device)
                y_predict = self.model(X)
                loss = loss_fn(y_predict, y)
                opt.zero_grad()
                loss.backward()
                opt.step()
            print(f'End of epoch {e} | current loss {loss.data}')

    def predict(self, X):
        return self.model(X)

    def predict_proba(self, X, target):
        pass

    def fit_predict(self, X, y, X_test=None, **kwargs):
        self.fit(X, y, **kwargs)

        if X_test is None:
            X_test = X

        return self.predict(X_test)


# Modify this if you want to use networks with other structures.
class MultiClassNet(nn.Module):
    def __init__(self,
                 in_d,
                 out_d,
                 hidden_d1=100,
                 hidden_d2=128,
                 hidden_d3=64):
        super().__init__()
        self.in_d = in_d
        self.out_d = out_d
        self.fc1 = nn.Linear(in_d, hidden_d1)
        self.fc2 = nn.Linear(hidden_d1, hidden_d2)
        self.fc3 = nn.Linear(hidden_d2, hidden_d3)
        self.fc4 = nn.Linear(hidden_d3, out_d)

    def fowrad(self, x):
        x = torch.flatten(x, 1)
        x = self.fc1(x)
        x = F.relu(x)
        x = self.fc2(x)
        x = F.relu(x)
        x = self.fc3(x)
        output = F.relu(x)
        return output


class MCNWrapper(MLModel):
    def __init__(self, treatment_net):
        super().__init__()
        self.model = treatment_net

    def fit(self, X, y, **optim_config):
        loss = nn.CrossEntropyLoss()
        self._fit_nn_torch(X, y, loss=loss, **optim_config)

    def predict(self, X, label=True):
        y_pred = nn.Softmax(dim=1)(self.model(X))
        if label:
            y_pred = y_pred.argmax(dim=1).view(X.shape[0], -1)
        return y_pred

    def predict_proba(self, X, target=None):
        y_pred = self.predict(X, label=False)
        if target is None:
            return y_pred
        else:
            return y_pred[:, target]

    def sample(self, X, sample_num):
        # This method is unnecessary for many tasks.
        pass



# -*- coding:utf-8 -*-
"""

"""
import copy

import numpy as np
import pandas as pd
from sklearn.compose import make_column_selector

from ylearn.utils import logging

logger = logging.get_logger(__name__)

column_object = make_column_selector(dtype_include=['object'])
column_object_category_bool_int = make_column_selector(
    dtype_include=['object', 'category', 'bool',
                   'int', 'int8', 'int16', 'int32', 'int64',
                   'uint', 'uint8', 'uint16', 'uint32', 'uint64'])
column_int = make_column_selector(
    dtype_include=['int', 'int8', 'int16', 'int32', 'int64',
                   'uint', 'uint8', 'uint16', 'uint32', 'uint64'])


class _CleanerHelper:
    @staticmethod
    def reduce_mem_usage(df, excludes=None):
        """
        Adaption from :https://blog.csdn.net/xckkcxxck/article/details/88170281
        """
        numerics = ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']
        start_mem = df.memory_usage().sum() / 1024 ** 2
        for col in df.columns:
            if excludes is not None and col in excludes:
                continue
            col_type = df[col].dtypes
            if col_type in numerics:
                c_min = df[col].min()
                c_max = df[col].max()
                if str(col_type)[:3] == 'int':
                    if c_min > np.iinfo(np.int8).min and c_max < np.iinfo(np.int8).max:
                        df[col] = df[col].astype(np.int8)
                    elif c_min > np.iinfo(np.int16).min and c_max < np.iinfo(np.int16).max:
                        df[col] = df[col].astype(np.int16)
                    elif c_min > np.iinfo(np.int32).min and c_max < np.iinfo(np.int32).max:
                        df[col] = df[col].astype(np.int32)
                    elif c_min > np.iinfo(np.int64).min and c_max < np.iinfo(np.int64).max:
                        df[col] = df[col].astype(np.int64)
                else:
                    if c_min > np.finfo(np.float32).min and c_max < np.finfo(np.float32).max:
                        df[col] = df[col].astype(np.float32)
                    else:
                        df[col] = df[col].astype(np.float64)
        end_mem = df.memory_usage().sum() / 1024 ** 2
        if logger.is_info_enabled():
            logger.info('Mem. usage decreased to {:5.2f} Mb ({:.1f}% reduction)'
                        .format(end_mem, 100 * (start_mem - end_mem) / start_mem))

    @staticmethod
    def replace_nan_chars(X, nan_chars):
        return X.replace(nan_chars, np.nan)

    def correct_object_dtype(self, X, df_meta=None, excludes=None):
        if df_meta is None:
            Xt = X[[c for c in X.columns.to_list() if c not in excludes]] if excludes else X
            cats = []

            cons = [c for c in column_object(Xt) if c not in cats]
            df_meta = {'object': cats, 'float': cons}

        X = self._correct_object_dtype_as(X, df_meta)

        return X

    def _correct_object_dtype_as(self, X, df_meta):
        for dtype, columns in df_meta.items():
            columns = [c for c in columns if str(X[c].dtype) != dtype]
            if len(columns) == 0:
                continue

            if dtype in ['object', 'str']:
                X[columns] = X[columns].astype(dtype)
            else:
                for col in columns:
                    try:
                        if str(X[col].dtype) != str(dtype):
                            X[col] = X[col].astype(dtype)
                    except Exception as e:
                        if logger.is_debug_enabled():
                            logger.debug(f'Correct object column [{col}] as {dtype} failed. {e}')

        return X

    def drop_duplicated_columns(self, X, excludes=None):
        duplicates = self._get_duplicated_columns(X)
        dup_cols = [i for i, v in duplicates.items() if v and (excludes is None or i not in excludes)]
        if len(dup_cols) > 0:
            columns = [c for c in X.columns.to_list() if c not in dup_cols]
            X = X[columns]

        return X, dup_cols

    def drop_constant_columns(self, X, excludes=None):
        nunique = self._get_df_uniques(X)
        const_cols = [i for i, v in nunique.items() if v <= 1 and (excludes is None or i not in excludes)]
        if len(const_cols) > 0:
            columns = [c for c in X.columns.to_list() if c not in const_cols]
            X = X[columns]
        return X, const_cols

    def drop_idness_columns(self, X, excludes=None):
        cols = column_object_category_bool_int(X)
        if len(cols) <= 0:
            return X, []

        X_ = X[cols]
        nunique = self._get_df_uniques(X_)
        rows = len(X)
        threshold = 0.99  # cfg.idness_threshold
        dropped = [c for c, n in nunique.items() if n / rows > threshold and (excludes is None or c not in excludes)]
        if len(dropped) > 0:
            columns = [c for c in X.columns.to_list() if c not in dropped]
            X = X[columns]
        return X, dropped

    @staticmethod
    def _get_df_uniques(df):
        return df.nunique(dropna=True)

    @staticmethod
    def _get_duplicated_columns(df):
        return df.T.duplicated()


class DataCleaner:
    def __init__(self, nan_chars=None, correct_object_dtype=True, drop_constant_columns=True,
                 drop_duplicated_columns=False, drop_label_nan_rows=True, drop_idness_columns=True,
                 replace_inf_values=np.nan, drop_columns=None, reserve_columns=None,
                 reduce_mem_usage=False, int_convert_to='float'):
        self.nan_chars = nan_chars
        self.correct_object_dtype = correct_object_dtype
        self.drop_constant_columns = drop_constant_columns
        self.drop_label_nan_rows = drop_label_nan_rows
        self.drop_idness_columns = drop_idness_columns
        self.replace_inf_values = replace_inf_values
        self.drop_columns = drop_columns
        self.reserve_columns = reserve_columns
        self.drop_duplicated_columns = drop_duplicated_columns
        self.reduce_mem_usage = reduce_mem_usage
        self.int_convert_to = int_convert_to

        # fitted
        self.df_meta_ = None
        self.columns_ = None
        self.dropped_constant_columns_ = None
        self.dropped_idness_columns_ = None
        self.dropped_duplicated_columns_ = None

    def get_params(self):
        return {
            'nan_chars': self.nan_chars,
            'correct_object_dtype': self.correct_object_dtype,
            'drop_constant_columns': self.drop_constant_columns,
            'drop_label_nan_rows': self.drop_label_nan_rows,
            'drop_idness_columns': self.drop_idness_columns,
            # 'replace_inf_values': self.replace_inf_values,
            'drop_columns': self.drop_columns,
            'reserve_columns': self.reserve_columns,
            'drop_duplicated_columns': self.drop_duplicated_columns,
            'reduce_mem_usage': self.reduce_mem_usage,
            'int_convert_to': self.int_convert_to
        }

    @staticmethod
    def get_helper(X, y):
        return _CleanerHelper()

    @staticmethod
    def _drop_columns(X, cols):
        if cols is None or len(cols) <= 0:
            return X
        X = X[[c for c in X.columns.to_list() if c not in cols]]
        return X

    def clean_data(self, X, y, *, df_meta=None, reduce_mem_usage):
        y_name = '__tabular-toolbox__Y__'

        if y is not None:
            X[y_name] = y

        helper = self.get_helper(X, y)

        if self.nan_chars is not None:
            logger.debug(f'replace chars{self.nan_chars} to NaN')
            # X = X.replace(self.nan_chars, np.nan)
            X = helper.replace_nan_chars(X, self.nan_chars)

        if y is not None:
            if self.drop_label_nan_rows:
                logger.debug('clean the rows which label is NaN')
                X = X.dropna(subset=[y_name])
            y = X.pop(y_name)

        if self.drop_columns is not None:
            logger.debug(f'drop columns:{self.drop_columns}')
            X = self._drop_columns(X, self.drop_columns)

        if self.drop_duplicated_columns:
            if self.dropped_duplicated_columns_ is not None:
                X = self._drop_columns(X, self.dropped_duplicated_columns_)
            else:
                X, self.dropped_duplicated_columns_ = helper.drop_duplicated_columns(X, self.reserve_columns)
            logger.info(f'drop duplicated columns: "{self.dropped_duplicated_columns_}')

        if self.drop_idness_columns:
            if self.dropped_idness_columns_ is not None:
                X = self._drop_columns(X, self.dropped_idness_columns_)
            else:
                X, self.dropped_idness_columns_ = helper.drop_idness_columns(X, self.reserve_columns)
            logger.debug(f'drop idness columns: {self.dropped_idness_columns_}')

        if self.drop_constant_columns:
            if self.dropped_constant_columns_ is not None:
                X = self._drop_columns(X, self.dropped_constant_columns_)
            else:
                X, self.dropped_constant_columns_ = helper.drop_constant_columns(X, self.reserve_columns)
            logger.debug(f'drop constant columns: {self.dropped_constant_columns_}')

        if self.replace_inf_values is not None:
            logger.info(f'replace [inf,-inf] to {self.replace_inf_values}')
            int_cols = column_int(X)
            if len(int_cols) > 0:
                columns = [c for c in X.columns.to_list() if c not in int_cols]
                if len(columns) > 0:
                    X[columns] = X[columns].replace([np.inf, -np.inf], self.replace_inf_values)
            else:
                X = X.replace([np.inf, -np.inf], self.replace_inf_values)

        if self.correct_object_dtype:
            logger.debug('correct data type for object columns.')
            # for col in column_object(X):
            #     try:
            #         X[col] = X[col].astype('float')
            #     except Exception as e:
            #         logger.error(f'Correct object column [{col}] failed. {e}')
            X = helper.correct_object_dtype(X, df_meta, excludes=self.reserve_columns)

        if self.int_convert_to is not None:
            int_cols = column_int(X)
            if self.reserve_columns:
                int_cols = list(filter(lambda _: _ not in self.reserve_columns, int_cols))
            if len(int_cols) > 0:
                logger.info(f'convert int type to {self.int_convert_to}, {int_cols}')
                X[int_cols] = X[int_cols].astype(self.int_convert_to)

        o_cols = column_object(X)
        if self.reserve_columns:
            o_cols = list(filter(lambda _: _ not in self.reserve_columns, o_cols))
        if o_cols:
            logger.info(f'convert object type to str, {o_cols}')
            X[o_cols] = X[o_cols].astype('str')

        if reduce_mem_usage:
            logger.info('try reduce memory usage')
            helper.reduce_mem_usage(X, excludes=self.reserve_columns)

        return X, y

    @staticmethod
    def _copy(X, y):
        if isinstance(X, pd.DataFrame):
            X = copy.deepcopy(X)
        else:
            X = X.copy()

        if y is not None:
            if isinstance(y, (np.ndarray, pd.Series, pd.DataFrame)):
                y = copy.deepcopy(y)
            else:
                y = y.copy()

        return X, y

    def fit_transform(self, X, y=None, copy_data=True):
        if copy_data:
            X, y = self._copy(X, y)

        X, y = self.clean_data(X, y, reduce_mem_usage=self.reduce_mem_usage)

        logger.debug('collect meta info from data')
        df_meta = {}
        for col_info in zip(X.columns.to_list(), X.dtypes):
            dtype = str(col_info[1])
            if df_meta.get(dtype) is None:
                df_meta[dtype] = []
            df_meta[dtype].append(col_info[0])

        logger.info(f'cleaned dataframe meta:{df_meta}')
        self.df_meta_ = df_meta
        self.columns_ = X.columns.to_list()

        return X, y

    def transform(self, X, y=None, copy_data=True):
        if copy_data:
            X, y = self._copy(X, y)

        orig_columns = X.columns.to_list()
        X, y = self.clean_data(X, y, df_meta=self.df_meta_, reduce_mem_usage=False)
        # if self.df_meta_ is not None:
        #     logger.debug('processing with meta info')
        #     all_cols = []
        #     for dtype, cols in self.df_meta_.items():
        #         all_cols += cols
        #         X[cols] = X[cols].astype(dtype)
        #     drop_cols = set(X.columns.to_list()) - set(all_cols)
        #     X = X[all_cols]
        #     logger.debug(f'droped columns:{drop_cols}')

        X = X[self.columns_]
        if logger.is_info_enabled():
            dropped = [c for c in orig_columns if c not in self.columns_]
            logger.info(f'drop columns: {dropped}')

        if y is None:
            return X
        else:
            return X, y

    def append_drop_columns(self, columns):
        if self.df_meta_ is None:
            if self.drop_columns is None:
                self.drop_columns = []
            self.drop_columns = list(set(self.drop_columns + columns))
        else:
            meta = {}
            for dtype, cols in self.df_meta_.items():
                meta[dtype] = [c for c in cols if c not in columns]
            self.df_meta_ = meta

        self.columns_ = [c for c in self.columns_ if c not in columns]

    def _repr_html_(self):
        cleaner_info = [
            ('Meta', self.df_meta_),
            ('Dropped constant columns', self.dropped_constant_columns_),
            ('Dropped idness columns', self.dropped_idness_columns_),
            ('Dropped duplicated columns', self.dropped_duplicated_columns_),
            ('-------------params-------------', '-------------values-------------'),
            ('nan_chars', self.nan_chars),
            ('correct_object_dtype', self.correct_object_dtype),
            ('drop_constant_columns', self.drop_constant_columns),
            ('drop_label_nan_rows', self.drop_label_nan_rows),
            ('drop_idness_columns', self.drop_idness_columns),
            ('replace_inf_values', self.replace_inf_values),
            ('drop_columns', self.drop_columns),
            ('reserve_columns', self.reserve_columns),
            ('drop_duplicated_columns', self.drop_duplicated_columns),
            ('reduce_mem_usage', self.reduce_mem_usage),
            ('int_convert_to', self.int_convert_to),
        ]

        html = pd.DataFrame(cleaner_info, columns=['key', 'value'])._repr_html_()
        return html
